import threading
from collections import defaultdict
from enum import Enum
from typing import Any, Callable, List, Optional, TYPE_CHECKING, Union, Iterable

import numpy as np
import pandas as pd

forbidden_symbols = ':/|\\?*<>"'

if TYPE_CHECKING:
    from pandas import DataFrame


class OrEvent(threading.Event):
    @staticmethod
    def orify(event, changed_callback):
        def or_clear(self):
            self._clear()
            self.changed()

        def or_set(self):
            self._set()
            self.changed()

        event._set = event.set
        event._clear = event.clear
        event.changed = changed_callback
        event.set = lambda: or_set(event)
        event.clear = lambda: or_clear(event)

    def __init__(self, *events) -> None:
        super().__init__()

        def changed():
            bools = [ev.is_set() for ev in events]
            if any(bools):
                self.set()
            else:
                self.clear()

        for event in events:
            OrEvent.orify(event, changed)

        # initialize
        changed()


class RepeatedTimer(threading.Timer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.daemon = True

    def not_finished(self) -> bool:
        return not self.finished.is_set()

    def run(self):
        while self.not_finished():
            self.finished.wait(self.interval)
            if not self.finished.is_set():
                self.function(*self.args, **self.kwargs)


class CallbackHandler:
    def __init__(self, max_listeners: int = 1):
        self._lock = threading.Lock()
        self._max_listeners = max_listeners
        self._callbacks = defaultdict(list)

    def on(self, event: str, callback: Callable) -> None:
        with self._lock:
            if len(self._callbacks[event]) >= self._max_listeners:
                self._callbacks[event].pop(0)
            self._callbacks[event].append(callback)

    def remove_listener(self, event: str, callback: Callable) -> None:
        with self._lock:
            self._callbacks[event].remove(callback)

    def emit(self, event: str, *args, **kwargs) -> None:
        # .copy() makes sure that underlying list will not change during iteration
        for callback in self._callbacks.get(event, []).copy():
            callback(*args, **kwargs)

    def remove_all_listeners(self, event: Optional[str] = None) -> None:
        with self._lock:
            if event is not None:
                self._callbacks[event] = []
            else:
                self._callbacks = defaultdict(list)


def ohlc(self, interval: str = "1min", *args, **kwargs) -> "DataFrame":
    call_from_recorder = kwargs.pop("call_from_recorder", False)
    df = self.fillna(method="ffill")
    df.fillna(method="bfill", inplace=True)
    df = df.astype("float")

    df = df.resample(interval, *args, **kwargs).ohlc()

    if call_from_recorder:
        df.fillna(pd.NA, inplace=True)
    return df


def get_from_path(obj, path, delim="."):
    if isinstance(path, list):
        splitted = path

    else:
        splitted = path.split(delim)

    for k in splitted:
        if hasattr(obj, "get"):
            obj = obj.get(k)

        elif iterable(obj) and is_int(k):
            try:
                obj = obj[int(k)]
            except (IndexError, KeyError):
                return None

    return obj


def is_int(obj):
    if isinstance(obj, str):
        try:
            int(obj)
        except Exception:
            return False
        else:
            return True
    return isinstance(obj, int)


def iterable(obj):
    if isinstance(obj, str):
        return False
    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True


def bool_or_none_to_str(value: Any) -> Union[str, Any]:
    """
    Coerce a bool type or None type into a string value.

    Note that we prefer JSON-style 'true'/'false' for boolean values here.
    """
    if isinstance(value, bool):
        return str(value).lower()
    elif value is None:
        return ""

    return value


def urljoin(*pieces):
    # first piece have a leading slash
    if pieces and len(pieces[0]) > 1 and pieces[0][0] == "/":
        pieces = ("/",) + pieces
    # last piece have a trailing slash
    if pieces and len(pieces[-1]) > 1 and pieces[-1][-1] == "/":
        pieces = pieces + ("/",)
    return "/".join(s.strip("/") for s in pieces)


def is_any_defined(*args):
    return any(args)


def is_all_defined(*args):
    return all(args)


def is_all_same_type(item_type, items):
    return all(isinstance(item, item_type) for item in items)


def make_counter():
    i = 0

    def counter():
        nonlocal i
        i += 1
        return i

    return counter


def get_response_reason(response):
    if hasattr(response, "reason_phrase"):
        return response.reason_phrase
    elif hasattr(response, "reason"):
        return response.reason
    return "unknown reason"


class cached_property(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls=None):
        result = instance.__dict__[self.func.__name__] = self.func(instance)
        return result


def is_tuple(what):
    return type(what) == tuple


def is_all_str(input_list: list) -> bool:
    return all([isinstance(i, str) for i in input_list])


def validate_bool_value(value: bool) -> bool:
    if isinstance(value, bool):
        return value
    else:
        raise ValueError(
            f"Please provide boolean value ('True' or 'False'), "
            f"current value is '{value}'"
        )


def args_parser(param: Any, expect_none: bool = False) -> Optional[list]:
    if isinstance(param, str):
        return [param]

    if isinstance(param, list):
        if is_all_str(param):
            return param
        else:
            raise ValueError(f"Not all elements are strings in {param}")

    if param is None and expect_none:
        return None

    raise ValueError(f"Invalid type, expected str or list:\n {type(param)} is given")


class ArgsParser:
    def __init__(self, parse) -> None:
        self.parse = parse

    def get_str(self, *args, delim=None) -> str:
        if delim is not None:
            retval = delim.join(str(item) for item in self.get_list(*args))
        else:
            retval = self.parse(*args)
            if not isinstance(retval, str):
                retval = str(retval)
        return retval

    def get_list(self, *args) -> list:
        retval = self.parse(*args)
        if not isinstance(retval, list):
            retval = [retval]
        return retval

    def get_float(self, *args) -> float:
        retval = self.parse(*args)
        if isinstance(retval, np.datetime64):
            retval = retval.astype(float)
        else:
            retval = float(retval)
        return retval

    def get_bool(self, *args) -> bool:
        retval = self.parse(*args)
        if not isinstance(retval, bool):
            retval = bool(retval)
        return retval

    def get_unique(self, *args) -> list:
        return list(dict.fromkeys(self.get_list(*args)))


class EnumArgsParser(ArgsParser):
    def __init__(self, parse, parse_to_enum) -> None:
        super().__init__(parse)
        self.parse_to_enum = parse_to_enum

    def get_enum(self, *args) -> Enum:
        return self.parse_to_enum(*args)


def parse_list_of_str(param: Union[str, list]) -> list:
    if isinstance(param, str):
        return [param]

    if isinstance(param, list):
        if is_all_same_type(str, param):
            return param
        else:
            raise ValueError(f"Not all elements are strings in {param}")

    raise TypeError(
        f"Invalid type, expected str or list, {type(param).__name__} is given"
    )


def parse_iterable_of_str(param: Union[str, Iterable[str]]) -> list:
    if isinstance(param, str):
        return [param]

    if isinstance(param, Iterable):
        if not isinstance(param, list):
            param = list(param)
        if is_all_same_type(str, param):
            return param
        else:
            raise ValueError(f"Not all elements are strings in {param}")

    raise TypeError(
        f"Invalid type, expected str or list, {type(param).__name__} is given"
    )


def make_parse_enum(enum_class: iterable, can_be_lower: bool = True) -> callable:
    enum_values = [k.value for k in enum_class]

    def parse_enum(param: Union[str, list, Enum]) -> Union[str, list]:
        if isinstance(param, list):
            return [parse_enum(p) for p in param]

        if isinstance(param, enum_class):
            return param.value

        if param in enum_values:
            return param

        if can_be_lower and hasattr(param, "upper"):
            param_upper = param.upper()
            upper_enum_values = {value.value.upper(): value for value in enum_class}

            if param_upper in upper_enum_values:
                return upper_enum_values[param_upper].value

        raise AttributeError(f"Value '{param}' must be in {enum_values}")

    return parse_enum


def make_convert_to_enum(enum_class: iterable) -> callable:
    lower_values = {item.value.lower(): item for item in enum_class}

    def convert_to_enum(
        param: Union[str, List[str], Enum, List[Enum]]
    ) -> Union[Enum, List[Enum]]:
        if isinstance(param, str) and param.lower() in lower_values:
            return lower_values[param.lower()]

        if isinstance(param, list):
            return [convert_to_enum(p) for p in param]

        if isinstance(param, enum_class):
            return param

        if isinstance(param, str) and param.upper() in enum_class.__members__.keys():
            param = enum_class.__members__[param.upper()]
            return param

        raise ValueError(f"Cannot convert param '{param}'")

    return convert_to_enum


def make_callback(func: Callable) -> Callable:
    """Return a callback function with correct arguments order.

    For user defined callback functions in OMMStream, RDPStream and trade data streams,
    correct arguments order is:
        Callable[[dict, "Stream"], Any]

    For user defined callback functions in pricing streams correct arguments order is:
        Callable[[Any, str, "Stream"], Any]

    For user defined callback in financial contracts correct arguments order is:
        Callable[[list, "Stream"], Any]

    Uses in OMMStream, RDPStream, financial contracts,  pricing and trade data streams.
    """

    def callback(*args):
        args = reversed(args)
        func(*args)

    return callback


def parse_hp_universe(universe: Union[str, list]) -> list:
    universe = parse_list_of_str(universe)
    universe = [i for i in universe if i not in {"", " "}]
    if universe:
        return universe
    raise ValueError("List of universes is empty, nothing to process")


def validate_types(val: Any, types: Union[list, tuple], val_name: str = "") -> None:
    t_names = [tp.__name__ for tp in types if tp if tp.__name__ != "NoneType"]

    if None in types:
        raise TypeError("Use 'type(None)' instead 'None', in 'types'")

    if type(val) not in types:
        raise TypeError(
            f"Parameter '{val_name}' of invalid type provided: '{type(val).__name__}', "
            f"expected types: {t_names}"
        )


fields_arg_parser = ArgsParser(parse_list_of_str)
universe_arg_parser = fields_arg_parser
hp_universe_parser = ArgsParser(parse_hp_universe)
custom_insts_historical_universe_parser = ArgsParser(parse_hp_universe)
iterator_str_arg_parser = ArgsParser(parse_iterable_of_str)
attributes_arg_parser = fields_arg_parser


def make_enum_arg_parser(*args, **kwargs):
    return ArgsParser(make_parse_enum(*args, **kwargs))


def make_convert_to_enum_arg_parser(*args, **kwargs):
    return ArgsParser(make_convert_to_enum(*args, **kwargs))


def make_enum_arg_parser_by_members(enum_class, can_be_lower=True):
    parser = make_parse_enum(enum_class, can_be_lower)

    def _parser(value):
        lower_values = {
            name.lower(): item for name, item in enum_class.__members__.items()
        }
        if isinstance(value, str) and value.lower() in lower_values:
            value = lower_values[value.lower()]
        return parser(value)

    return ArgsParser(_parser)


def merge_dict_to_dict(dest: dict, source: dict) -> dict:
    for source_key, source_value in source.items():
        dest_value = dest.get(source_key)

        if (
            source_key in dest
            and isinstance(source_value, dict)
            and isinstance(dest_value, dict)
        ):
            merge_dict_to_dict(dest_value, source_value)

        else:
            dest[source_key] = source_value

    return dest


def extend_params(params, extended_params):
    if extended_params:
        # params -> [("param1", "val1"), ]
        params = dict(params)
        # result -> {"param1": "val1"}
        params.update(extended_params)
        # result -> {"param1": "val1", "extended_param": "value"}
        # return [("param1", "val1"), ("extended_param", "value")]
        return list(params.items())
    return params


def get_correct_filename(filename, replace_with=""):
    result = filename
    for symbol in forbidden_symbols:
        result = result.replace(symbol, replace_with)
    return result



import pytest
from refinitiv.data.content.ownership.insider import (
    shareholders_report,
    transaction_report,
)


def test_shareholders_report():
    try:
        shareholders_report.Definition(universe="")
    except Exception as e:
        assert False, str(e)


def test_transaction_report():
    try:
        transaction_report.Definition(universe="")
    except Exception as e:
        assert False, str(e)



from typing import Optional, TYPE_CHECKING

from ._valuation_time import ValuationTime
from ...._enums._extrapolation_mode import ExtrapolationMode
from ...._enums._interpolation_mode import InterpolationMode
from ...._object_definition import ObjectDefinition


if TYPE_CHECKING:
    from ......_types import OptStr, OptBool


class FxForwardCurveParameters(ObjectDefinition):
    """
    Generates the Cross Currency curves for the definitions provided

    Parameters
    ----------
    extrapolation_mode : ExtrapolationMode, optional

    interpolation_mode : InterpolationMode, optional

    valuation_time : ValuationTime, optional
        The time identified by offsets at which the zero coupon curve is generated.
    ignore_invalid_instrument : bool, optional
        Ignore invalid instrument to calculate the curve.  if false and some instrument
        are invlide, the curve is not calculated and an error is returned.  the default
        value is 'true'.
    ignore_pivot_currency_holidays : bool, optional

    use_delayed_data_if_denied : bool, optional
        Use delayed ric to retrieve data when not permissioned on constituent ric. the
        default value is 'false'.
    valuation_date : str, optional

    valuation_date_time : str, optional
        The date and time at which the zero coupon curve is generated. the value is
        expressed in iso 8601 format: yyyy-mm-ddt00:00:00z (e.g., '2021-01-01t14:00:00z'
        or '2021-01-01t14:00:00+02:00'). only one parameter of valuationdate and
        valuationdatetime must be specified.
    """

    def __init__(
        self,
        extrapolation_mode: Optional[ExtrapolationMode] = None,
        interpolation_mode: Optional[InterpolationMode] = None,
        valuation_time: Optional[ValuationTime] = None,
        ignore_invalid_instrument: "OptBool" = None,
        ignore_pivot_currency_holidays: "OptBool" = None,
        use_delayed_data_if_denied: "OptBool" = None,
        valuation_date: "OptStr" = None,
        valuation_date_time: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.extrapolation_mode = extrapolation_mode
        self.interpolation_mode = interpolation_mode
        self.valuation_time = valuation_time
        self.ignore_invalid_instrument = ignore_invalid_instrument
        self.ignore_pivot_currency_holidays = ignore_pivot_currency_holidays
        self.use_delayed_data_if_denied = use_delayed_data_if_denied
        self.valuation_date = valuation_date
        self.valuation_date_time = valuation_date_time

    @property
    def extrapolation_mode(self):
        """
        :return: enum ExtrapolationMode
        """
        return self._get_enum_parameter(ExtrapolationMode, "extrapolationMode")

    @extrapolation_mode.setter
    def extrapolation_mode(self, value):
        self._set_enum_parameter(ExtrapolationMode, "extrapolationMode", value)

    @property
    def interpolation_mode(self):
        """
        :return: enum InterpolationMode
        """
        return self._get_enum_parameter(InterpolationMode, "interpolationMode")

    @interpolation_mode.setter
    def interpolation_mode(self, value):
        self._set_enum_parameter(InterpolationMode, "interpolationMode", value)

    @property
    def valuation_time(self):
        """
        The time identified by offsets at which the zero coupon curve is generated.
        :return: object ValuationTime
        """
        return self._get_object_parameter(ValuationTime, "valuationTime")

    @valuation_time.setter
    def valuation_time(self, value):
        self._set_object_parameter(ValuationTime, "valuationTime", value)

    @property
    def ignore_invalid_instrument(self):
        """
        Ignore invalid instrument to calculate the curve.  if false and some instrument
        are invlide, the curve is not calculated and an error is returned.  the default
        value is 'true'.
        :return: bool
        """
        return self._get_parameter("ignoreInvalidInstrument")

    @ignore_invalid_instrument.setter
    def ignore_invalid_instrument(self, value):
        self._set_parameter("ignoreInvalidInstrument", value)

    @property
    def ignore_pivot_currency_holidays(self):
        """
        :return: bool
        """
        return self._get_parameter("ignorePivotCurrencyHolidays")

    @ignore_pivot_currency_holidays.setter
    def ignore_pivot_currency_holidays(self, value):
        self._set_parameter("ignorePivotCurrencyHolidays", value)

    @property
    def use_delayed_data_if_denied(self):
        """
        Use delayed ric to retrieve data when not permissioned on constituent ric. the
        default value is 'false'.
        :return: bool
        """
        return self._get_parameter("useDelayedDataIfDenied")

    @use_delayed_data_if_denied.setter
    def use_delayed_data_if_denied(self, value):
        self._set_parameter("useDelayedDataIfDenied", value)

    @property
    def valuation_date(self):
        """
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)

    @property
    def valuation_date_time(self):
        """
        The date and time at which the zero coupon curve is generated. the value is
        expressed in iso 8601 format: yyyy-mm-ddt00:00:00z (e.g., '2021-01-01t14:00:00z'
        or '2021-01-01t14:00:00+02:00'). only one parameter of valuationdate and
        valuationdatetime must be specified.
        :return: str
        """
        return self._get_parameter("valuationDateTime")

    @valuation_date_time.setter
    def valuation_date_time(self, value):
        self._set_parameter("valuationDateTime", value)



# coding: utf8

from enum import Enum, unique


@unique
class IndexConvexityAdjustmentMethod(Enum):
    BLACK_SCHOLES = "BlackScholes"
    LINEAR_SWAP_MODEL = "LinearSwapModel"
    NONE = "None"
    REPLICATION = "Replication"



def check_all_packages_names_contain_keyword(response, keyword):
    for package in response.data.raw["value"]:
        assert (
            keyword.lower() in package["packageName"].lower()
        ), f"Found package which does not contain expected package name {keyword}: \n {package}"


def check_all_packages_types(response, package_type):
    for package in response.data.raw["value"]:
        assert (
            package_type.lower() == package["packageType"].lower()
        ), f"Found package which does not contain expected type {package_type}: \n {package}"


def check_bucket_name_in_all_packages(response, bucket_name):
    for package in response.data.raw["value"]:
        assert (
            bucket_name in package["bucketNames"]
        ), f"Found package which does not belong to expected bucket name {bucket_name}: \n {package}"


def check_included_total_result_in_response(response):
    assert (
        "totalResults" in response.data.raw.keys()
    ), f"Response does not contain totalResults property: \n{response}"


def check_included_entitlement_result_in_response(response):
    packages = response.data.raw["value"]
    for package in packages:
        assert (
            "isEntitled" in package.keys()
        ), f"Found package which does not have isEntitled property: {package}"


def check_packages_in_two_responses_are_not_the_same(response1, response2):
    for package in response1.data.raw["value"]:
        assert (
            package not in response2.data.raw["value"]
        ), f"Package from first request found in list of packages in the second request: \n {package}"


def check_packages_in_second_response_are_on_proper_position_in_first_response(
    response1, response2, page_size, page
):
    packages_first_chunk = response1.data.raw["value"]
    packages_second_chunk = response2.data.raw["value"]
    position_index = page_size * (page - 1)
    for package in packages_second_chunk:
        assert package == packages_first_chunk[position_index], (
            f"Package from requested page  does not match the package on position {position_index}. \n"
            f"Expected package:  {packages_first_chunk[position_index]}. \nRetrieved package: {package}"
        )
        position_index += 1



import allure
import pytest

import refinitiv.data as rd
from tests.integration.access.conftest import (
    check_column_names_is_exist_in_response_and_df_not_empty,
)
from tests.integration.discovery.universe_expanders.conftest import FCHI_CONSTITUENTS
from tests.integration.helpers import (
    check_dataframe_column_date_for_datetime_type,
    check_if_dataframe_is_not_none,
    check_response_value,
)

FIELDS = [
    "BID",
    "TR.RevenueMean.currency",
    "TR.RevenueMean",
    "TR.RevenueMean",
    "TR.RevenueMean.date",
]
PARAMETERS = {
    "SCale": 6,
    "SDate": 0,
    "EDate": -3,
    "FRQ": "FY",
    "Curn": "USD",
}


@allure.suite("FinCoder layer")
@allure.feature("FinCoder layer - get_data")
@allure.severity(allure.severity_level.CRITICAL)
@pytest.mark.underlying_content("datagrid")
class TestGetData:
    @allure.title("Get mix of pricing and fundamental data")
    @pytest.mark.parametrize(
        "universe,fields,expected_fields",
        [
            (
                ["GOOG.O", "VOD.L", "EUR="],
                FIELDS,
                [
                    "Instrument",
                    "BID",
                    "Date",
                    "Currency",
                    "Revenue - Mean",
                ],
            ),
            (
                ["USD=", "EUR="],
                ["BID", "TR.Revenue"],
                [
                    "Instrument",
                    "BID",
                ],
            ),
        ],
    )
    @pytest.mark.caseid("36976688")
    @pytest.mark.smoke
    def test_get_mix_pricing_and_fundamental_data(
        self, open_desktop_session, universe, fields, expected_fields
    ):
        response = rd.get_data(
            universe=universe,
            fields=fields,
            parameters=PARAMETERS,
        )

        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_response_value(response, "Instrument", universe)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Get fundamental data with parameters")
    @pytest.mark.parametrize(
        "universe,fields,expected_fields",
        [
            (
                ["GOOG.O", "AAPL.O"],
                ["TR.PriceTargetMean(Source=ThomsonReuters)", "TR.Revenue.date"],
                ["Instrument", "Date", "Price Target - Mean"],
            ),
            (
                ["GOOG.O", "AAPL.O"],
                ["TR.Revenue(SDate=0, EDate=-3).date", "TR.Revenue(SDate=0, EDate=-3)"],
                ["Instrument", "Date", "Revenue"],
            ),
        ],
    )
    @pytest.mark.caseid("36976693")
    def test_get_fundamental_data_with_parameters(
        self, open_desktop_session, universe, fields, expected_fields
    ):
        response = rd.get_data(
            universe=universe,
            fields=fields,
        )

        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_response_value(response, "Instrument", universe)

    @allure.title("Get fundamental fields history")
    @pytest.mark.parametrize(
        "expected_fields",
        [["Instrument", "Date", "Revenue", "Gross Profit"]],
    )
    @pytest.mark.caseid("36976694")
    def test_get_fundamental_fields_history(
        self, open_desktop_session, expected_fields
    ):
        response = rd.get_data(
            universe=["GOOG.O", "AAPL.O"],
            fields=["TR.Revenue.date", "TR.Revenue", "TR.GrossProfit"],
        )

        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )

    @allure.title("Get fundamental data with one of the invalid universe")
    @pytest.mark.parametrize(
        "expected_fields",
        [["Instrument", "Date", "Revenue", "Gross Profit"]],
    )
    @pytest.mark.caseid("36976695")
    def test_get_fundamental_fields_data_with_one_of_the_invalid_universe(
        self, open_desktop_session, expected_fields
    ):
        response = rd.get_data(
            universe=["GOOG.O", "INVALID"],
            fields=["TR.Revenue.date", "TR.Revenue", "TR.GrossProfit"],
        )

        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )

    @allure.title("Get fundamental data with one invalid universe")
    @pytest.mark.caseid("36976696")
    def test_get_fundamental_fields_data_with_one_invalid_universe(
        self, open_desktop_session
    ):
        response = rd.get_data(
            universe="INVALID",
            fields=["TR.Revenue.date", "TR.Revenue", "TR.GrossProfit"],
        )

        assert response.empty, f"Response is not empty: {response}"

    @allure.title("Get fundamental data only with universe")
    @pytest.mark.parametrize(
        "universe,fields",
        [(["GOOG.O", "VOD.L"], None), ("GOOG.O", [])],
    )
    @pytest.mark.caseid("37039589")
    def test_get_fundamental_data_only_with_universe(
        self, open_desktop_session, universe, fields
    ):
        response = rd.get_data(universe=universe, fields=fields)

        assert not response.empty, f"{response}"
        check_response_value(response, "Instrument", universe)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title(
        "Get mix of pricing and fundamental data with parameter use_field_names_in_headers"
    )
    @pytest.mark.parametrize(
        "universe,expected_fields_udf,expected_fields_rdp",
        [
            (
                ["GOOG.O", "VOD.L", "EUR="],
                [
                    "Instrument",
                    "BID",
                    "TR.REVENUEMEAN.currency",
                    "TR.REVENUEMEAN",
                    "TR.REVENUEMEAN.DATE",
                ],
                [
                    "Instrument",
                    "BID",
                    "TR.RevenueMean",
                ],
            )
        ],
    )
    @pytest.mark.caseid("40099771")
    def test_get_mix_pricing_and_fundamental_data_with_use_field_names_in_headers(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        expected_fields_udf,
        expected_fields_rdp,
    ):
        response = rd.get_data(
            universe=universe,
            fields=FIELDS,
            parameters=PARAMETERS,
            use_field_names_in_headers=True,
        )
        expected_fields = expected_fields_udf
        if set_underlying_platform_config == "rdp-underlying-platform":
            expected_fields = expected_fields_rdp
        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_response_value(response, "Instrument", universe)

    @allure.title("Getting data by function and chain")
    @pytest.mark.parametrize(
        "universe,fields,expected_universe",
        [
            pytest.param(
                "0#.FCHI",
                ["BID", "ASK", "TR.Revenue"],
                FCHI_CONSTITUENTS,
                id="discovery_iterable",
            ),
            (
                'Peers("VOD.L")',
                ["BID", "ASK", "TR.Revenue"],
                ["LBTYA.OQ", "VTWRn.F", "TELIA.ST"],
            ),
            (
                'SCREEN(U(IN(Equity(active,public,primary))/*UNV:Public*/), IN(TR.HQCountryCode,"AR"), IN(TR.GICSIndustryCode,"401010"))',
                [
                    "TR.CommonName",
                    "TR.HeadquartersCountry",
                    "TR.GICSSector",
                    "TR.OrganizationStatusCode",
                    "TR.Revenue",
                    "BID",
                ],
                ["BMA.BA", "BBAR.BA", "BHIP.BA"],
            ),
            (
                "VOD.L",
                [
                    "TR.PeersRank",
                    "TR.RIC",
                    "AVAIL(TR.GrossProfit(Period=LTM,Methodology=InterimSum)/TR.TotalRevenue(Period=LTM,Methodology=InterimSum),TR.GrossProfit(Period=FY0)/TR.TotalRevenue(Period=FY0))*100",
                ],
                ["VOD.L"],
            ),
        ],
    )
    @pytest.mark.caseid("C43844230")
    @pytest.mark.smoke
    def test_get_data_by_function_and_chain(
        self,
        request,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
        expected_universe,
    ):
        if "discovery_iterable" in request.node.callspec.id:
            chain = rd.discovery.Chain(name=universe)
            list(chain.constituents)
            universe = chain

        response = rd.get_data(
            universe=universe,
            fields=fields,
        )

        list_of_instruments = list(response["Instrument"])
        check_if_dataframe_is_not_none(response)

        if expected_universe:
            for universe in expected_universe:
                assert (
                    universe in list_of_instruments
                ), f"The {universe} is missing in requested instruments"

    @allure.title("Getting data by function and chain without ADC fields")
    @pytest.mark.parametrize(
        "universe,fields",
        [
            ("0#.FCHI", ["BID", "ASK"]),
            ('Peers("VOD.L")', None),
            (
                'SCREEN(U(IN(Equity(active,public,primary))/*UNV:Public*/), IN(TR.HQCountryCode,"AR"), IN(TR.GICSIndustryCode,"401010"))',
                [
                    "ASK" "BID",
                ],
            ),
        ],
    )
    @pytest.mark.caseid("C54271593")
    def test_get_data_by_function_and_chain_without_adc_fields(
        self, open_session, universe, fields
    ):
        response = rd.get_data(
            universe=universe,
            fields=fields,
        )

        assert response.empty, f"Dataframe is not empty"

    @allure.title("Getting data with SCREEN which return empty response data from ADC")
    @pytest.mark.parametrize(
        "universe,fields",
        (
            [
                'SCREEN(U(IN(Equity(active,public,primary))/*UNV:Public*/), IN(TR.HQCountryCode,"AR"), IN(TR.GICSIndustryCode,"401020"))',
                ["TR.RIC"],
            ],
        ),
    )
    @pytest.mark.caseid("C43844230")
    def test_get_data_with_screen_which_return_empty_response(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
    ):
        response = rd.get_data(
            universe=universe,
            fields=fields,
        )

        assert response.empty, f"Response is not empty: {response}"

    @allure.title("Get Custom Instrument with mix of objects data")
    @pytest.mark.parametrize(
        "universes,fields,expected_fields",
        [
            ("S)universe", None, None),
            (None, ["TRDPRC_1", "INVALID"], ["Instrument", "TRDPRC_1", "INVALID"]),
            (
                ["IBM.N", "LSEG.L"],
                ["BID", "ASK"],
                ["Instrument", "BID", "ASK"],
            ),
        ],
        ids=["invalid", "partial_valid", "mix_objects"],
    )
    @pytest.mark.caseid("C43973494")
    def test_get_custom_instrument_with_mix_objects_data(
        self,
        open_desktop_session,
        create_instrument,
        universes,
        fields,
        expected_fields,
        request,
    ):
        if "partial_valid" in request.node.callspec.id:
            universes = [create_instrument(), create_instrument()]
        if "mix_objects" in request.node.callspec.id:
            universes += [create_instrument(), create_instrument()]

        response = rd.get_data(
            universe=universes,
            fields=fields,
        )
        if "invalid" in request.node.callspec.id:
            assert response.empty, f"Response is not empty: {response}"
        else:
            check_column_names_is_exist_in_response_and_df_not_empty(
                response, expected_fields
            )
            check_response_value(response, "Instrument", universes)



from typing import TYPE_CHECKING

from ._session_type import SessionType
from ._session import Session
from ._session_cxn_type import SessionCxnType

if TYPE_CHECKING:
    from ... import OpenState


class NullSession(Session):
    type = SessionType.NONE

    def __init__(self):
        Session.__init__(self, app_key="")

    def _get_session_cxn_type(self) -> SessionCxnType:
        return SessionCxnType.NONE

    def get_omm_login_message(self):
        return {}

    def get_rdp_login_message(self, stream_id):
        return {}

    async def open_async(self) -> "OpenState":
        # do nothing
        pass



import pytest

from refinitiv.data import _configure as configure

config_value = "value"

config = configure.get_config()


@pytest.mark.xdist_group(name="config_group")
def test_set_param_to_config():
    try:
        config["param-name"] = config_value
    except Exception as e:
        assert False, e


@pytest.mark.xdist_group(name="config_group")
def test_get_param_from_config():
    try:
        value = config["param-name"]
        assert value == config_value, value
    except Exception as e:
        assert False, e

    try:
        value = config.get_str("param-name")
        assert value == config_value, value
    except Exception as e:
        assert False, e



from typing import TYPE_CHECKING

from ._custom_instruments_data_provider import get_user_id
from ._stream_facade import Stream, _init_universe
from ..._tools import universe_arg_parser, fields_arg_parser, try_copy_to_list

if TYPE_CHECKING:
    from ..._types import OptStr, ExtendedParams, StrStrings, OptStrStrs


class Definition:
    """
    This class defines parameters for requesting events from custom instrument

    Parameters
    -----------
    universe : str
        The Id or Symbol of custom instrument to operate on. Use only for get_instrument().
    fields : str or list of str, optional
        Specifies the specific fields to be delivered when messages arrive
    api: str, optional
        Specifies the data source. It can be updated/added using config file
    extended_params : dict, optional
        If necessary other parameters

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments import Definition
    >>> definition = Definition(universe="S)MyNewInstrument")
    >>> stream = definition.get_stream()
    >>> stream.open()
    """

    def __init__(
        self,
        universe: "StrStrings",
        fields: "OptStrStrs" = None,
        api: "OptStr" = None,
        extended_params: "ExtendedParams" = None,
    ):
        extended_params = extended_params or {}
        universe = extended_params.pop("universe", universe)
        fields = extended_params.pop("fields", fields)
        fields = try_copy_to_list(fields)
        fields = fields_arg_parser.get_unique(fields or [])
        universe = try_copy_to_list(universe)
        self._universe = universe_arg_parser.get_list(universe)
        self._api = api
        self._fields = fields
        self._extended_params = extended_params

    def get_stream(self, session=None) -> Stream:
        uuid = get_user_id(session)
        universe = _init_universe(self._universe, session, uuid)
        stream = Stream(
            universe=universe,
            session=session,
            fields=self._fields,
            api=self._api,
            extended_params=self._extended_params,
            uuid=uuid,
        )
        return stream



# coding: utf8
from typing import List, TYPE_CHECKING, Callable, Any

import numpy as np
import pandas as pd
from numpy import iterable

from ._models import Surface
from .._content_provider import CurvesAndSurfacesRequestFactory, get_type_by_axis
from .._enums import Axis
from .._ipa_content_validator import IPAContentValidator
from ..._content_data_provider import ContentDataProvider
from ..._error_parser import ErrorParser
from ...._tools import cached_property
from ....delivery._data._data_provider import ValidatorContainer
from ....delivery._data._endpoint_data import EndpointData
from ....delivery._data._response_factory import ResponseFactory

if TYPE_CHECKING:
    from ....delivery._data._data_provider import ParsedData


# ---------------------------------------------------------------------------
#   ContentValidator
# ---------------------------------------------------------------------------


class SurfacesContentValidator(IPAContentValidator):
    @classmethod
    def content_data_status_is_not_error(cls, data: "ParsedData") -> bool:
        content_data = data.content_data
        if (
            isinstance(content_data.get("data"), list)
            and content_data.get("status") == "Error"
        ):
            data.error_codes = content_data.get("code")
            data.error_messages = content_data.get("message")
            return False

        return True

    @cached_property
    def validators(self) -> List[Callable[["ParsedData"], bool]]:
        return [
            self.content_data_is_not_none,
            self.content_data_status_is_not_error,
            self.any_element_have_no_error,
        ]


# ---------------------------------------------------------------------------
#   Data
# ---------------------------------------------------------------------------


def parse_axis(
    universe: dict,
    x_axis: Axis,
    y_axis: Axis,
) -> (np.array, np.array, np.array):
    """
      This method parsing the surface into lists row, column and matrix

      >>> from refinitiv.data.content import ipa
      >>> definition = ipa.surfaces.eti.Definition(
      ...     underlying_definition=ipa.surfaces.eti.EtiSurfaceDefinition(
      ...         instrument_code="BNPP.PA@RIC"
      ...     ),
      ...     surface_parameters=ipa.surfaces.eti.EtiCalculationParams(
      ...         price_side=ipa.surfaces.eti.PriceSide.MID,
      ...         volatility_model=ipa.surfaces.eti.VolatilityModel.SVI,
      ...         x_axis=ipa.surfaces.eti.Axis.STRIKE,
      ...         y_axis=ipa.surfaces.eti.Axis.DATE,
      ...     ),
      ...     surface_tag="1",
      ...     surface_layout=ipa.surfaces.eti.SurfaceLayout(
      ...         format=ipa.surfaces.eti.Format.MATRIX, y_point_count=10
      ...     ),
      ... )

      This example for surface_parameters with
      x_axis = Axis.STRIKE and y_axis = Axis.DATE

      |--→ column=Y
      ↓
    row=X

      >>> surface = universe.get("surface")
      >>> surface
      ... [
      ...   [None,    '2021-08-20', '2021-09-17', '2021-12-17', '2022-03-18'],
      ...   ['25.36',  63.76680855, 76.566676686, 514160483847, 45.563136028],
      ...   ['30.432', 56.20802369, 64.051912234, 46.118622487, 41.540289743],
      ...   ['35.504', 49.91436068, 51.916645386, 41.495311424, 37.870408673],
      ... ]

      Parameters
      ----------
      universe : dict
          dict with surface
      x_axis : Axis

      y_axis : Axis

      Returns
      -------
      (np.array, np.array, np.array)
          row, column, matrix or x, y, z

      Raises
      -------
      ValueError
          If x_axis or y_axis not correct
    """

    if not x_axis or not y_axis:
        raise ValueError(
            f"Cannot parse surface "
            f"without information about x_axis={x_axis} or y_axis={y_axis}"
        )

    surface = universe.get("surface")

    # column is ['2021-08-20', '2021-09-17', '2021-12-17', '2022-03-18', '2022-06-17']
    column = surface[0][1:]
    column = np.array(column, dtype=get_type_by_axis(y_axis))

    row = []
    matrix = []

    surface = surface[1:]
    # curve is ['25.36',  63.76680855, 76.566676686, 514160483847, 41.187204258]
    for curve in surface:
        # row is '25.36'
        row.append(curve[0])
        # matrix is [63.76680855, 76.566676686, 514160483847, 41.187204258]
        matrix.append(curve[1:])

    row = np.array(row, dtype=get_type_by_axis(x_axis))
    matrix = np.array(matrix, dtype=float)

    return row, column, matrix


def create_surfaces(raw, axes_params) -> List[Surface]:
    surfaces = []

    if raw and axes_params:
        for i, universe in enumerate(raw.get("data")):
            x_axis, y_axis = axes_params[i]
            row, column, matrix = parse_axis(universe, x_axis, y_axis)
            surface = Surface(row=row, column=column, matrix=matrix)
            surfaces.append(surface)

    return surfaces


class BaseData(EndpointData):
    _dataframe = None

    def __init__(self, raw, axes_params=None, **kwargs):
        super().__init__(raw, **kwargs)
        self._axes_params = axes_params

    @property
    def df(self):
        if self._dataframe is None and self._raw:
            data = self._raw.get("data")

            if data:
                data_frame = pd.DataFrame(data)
                data_frame.set_index("surfaceTag", inplace=True)

            else:
                data_frame = pd.DataFrame([])

            if not data_frame.empty:
                data_frame = data_frame.convert_dtypes()

            self._dataframe = data_frame

        return self._dataframe


class OneSurfaceData(BaseData):
    def __init__(self, raw, axes_params):
        super().__init__(raw, axes_params)
        self._surface = None

    @property
    def surface(self) -> Surface:
        if self._surface is None:
            surfaces = create_surfaces(self._raw, self._axes_params)
            self._surface = surfaces[0]
        return self._surface

    @property
    def df(self):
        if self._dataframe is None:
            data = {x: z for x, z in zip(self.surface.x, self.surface.z)}

            if data:
                data_frame = pd.DataFrame(data, index=self.surface.y)
            else:
                data_frame = super().df

            if not data_frame.empty:
                data_frame.fillna(pd.NA, inplace=True)
                data_frame = data_frame.convert_dtypes()

            self._dataframe = data_frame

        return self._dataframe


class SurfacesData(BaseData):
    def __init__(self, raw, axes_params):
        super().__init__(raw, axes_params)
        self._surfaces = None

    @property
    def surfaces(self) -> List[Surface]:
        if self._surfaces is None:
            self._surfaces = create_surfaces(self._raw, self._axes_params)
        return self._surfaces


# ---------------------------------------------------------------------------
#   ResponseFactory
# ---------------------------------------------------------------------------


def get_axis_params(obj, axis_params=None):
    if axis_params is None:
        axis_params = []

    if hasattr(obj, "_kwargs"):
        request_item = obj._kwargs.get("universe")

    else:
        request_item = obj

    surface_parameters = request_item.surface_parameters
    x_axis = surface_parameters._get_enum_parameter(Axis, "xAxis")
    y_axis = surface_parameters._get_enum_parameter(Axis, "yAxis")
    axis_params.append((x_axis, y_axis))
    return axis_params


class SurfaceResponseFactory(ResponseFactory):
    def create_data_success(self, parsed_data: "ParsedData", **kwargs) -> EndpointData:
        return self._do_create_data(self.get_raw(parsed_data), **kwargs)

    def create_data_fail(self, parsed_data: "ParsedData", **kwargs) -> EndpointData:
        return self._do_create_data({}, **kwargs)

    def _do_create_data(self, raw: Any, universe=None, **kwargs):
        if universe:
            if iterable(universe):
                axes_params = []
                for definition in universe:
                    get_axis_params(definition, axes_params)

                data = SurfacesData(raw, axes_params)

            else:
                axes_params = get_axis_params(universe)
                data = OneSurfaceData(raw, axes_params)

        else:
            data = BaseData(raw)

        return data


# ---------------------------------------------------------------------------
#   DataProvider
# ---------------------------------------------------------------------------

surfaces_data_provider = ContentDataProvider(
    request=CurvesAndSurfacesRequestFactory(),
    response=SurfaceResponseFactory(),
    validator=ValidatorContainer(content_validator=SurfacesContentValidator()),
    parser=ErrorParser(),
)



import logging
import os
from importlib import reload

import pytest

import refinitiv.data._log as session_logging
from refinitiv.data import _log as log
from refinitiv.data._core.session import DesktopSession
from .conftest import Record, stub_datetime
from ..conftest import args


@pytest.mark.parametrize(
    "input_log_level,expected_log_level",
    [
        ("", logging.INFO),
        ("DEBUG", logging.DEBUG),
        ("debug", logging.DEBUG),
        ("warn", logging.WARN),
        ("silent", logging.CRITICAL),
    ],
)
def test_convert_log_level(input_log_level, expected_log_level):
    # given

    # when
    testing_log_level = session_logging.convert_log_level(input_log_level)

    # then
    assert testing_log_level == expected_log_level


@pytest.mark.parametrize(
    "testing_data",
    [
        (Record("module:awesome"), "", False),
        (Record("module"), "*", True),
        (Record("session:module:a"), "session*", True),
        (Record("session:module:a"), "session:mod*", True),
        (Record("module:a:session"), "session*", False),
        (Record("module:a:session"), "module:a:*", True),
        (Record("module:b"), "session*, module:b, module:c", True),
        (Record("session:a"), "session*,module:b,module:c", True),
        (Record("module:a"), "*,-module:a", False),
        (Record("module:g"), "module:*, -module:b, -module:c", True),
        (Record("module:c"), "module:*, -module:b, -module:c", False),
        (Record("module:session"), "-module:*", False),
    ],
)
def test_make_filter(testing_data):
    # given
    record, filter_by, expected_result = testing_data

    # when
    filterer = log.make_filter(filter_by)
    testing_result = filterer(record)

    # then
    assert (
        testing_result == expected_result
    ), f"record={record.name}, filter_by={filter_by}"


def test_session_logger_and_logger_from_get_logger_is_same():
    # given
    session = DesktopSession("", name="test")
    expected_logger = session._logger

    # when
    testing_logger = log.get_logger(f"sessions.desktop.test.{session.session_id}")

    # then
    assert testing_logger == expected_logger


def test_get_logger_works_normal():
    try:
        logger = log.get_logger("test_get_logger")
    except Exception as e:
        assert False, e

    assert logger


def test_after_logger_disposed_handlers_are_empty():
    # given
    logger = log.get_logger("test_dispose_logger")

    # when
    log.dispose_logger(logger)

    # then
    assert logger.handlers == []


def test_dispose_logger_method_can_dispose_by_name():
    # given
    logger_name = "test_dispose_logger_name"
    logger = log.get_logger(logger_name)

    # when
    log.dispose_logger(logger_name)

    # then
    assert logger.handlers == []


def test_dispose_logger_method_can_dispose_by_logger_object():
    # given
    logger = log.get_logger("test_dispose_logger_name")

    # when
    log.dispose_logger(logger)

    # then
    assert logger.handlers == []


def test_existing_loggers_increase_after_call_create_logger_method():
    # given
    loggers_names = log.existing_loggers()
    before = len(loggers_names)

    # when
    log.create_logger("test_existing_loggers")
    loggers_names = log.existing_loggers()
    after = len(loggers_names)

    # then
    assert after - before == 1


def test_set_log_level_can_works_with_logger_name():
    # given
    logger_name = "test_set_log_level"
    logger = log.get_logger(logger_name)
    expected_value = 1
    input_value = 1

    # when
    log.set_log_level(logger_name, input_value)

    # then
    testing_value = logger.level
    assert testing_value == expected_value


def test_set_log_level_can_works_with_logger_object():
    # given
    logger = log.get_logger("test_set_log_level")
    expected_value = 5
    input_value = 5

    # when
    log.set_log_level(logger, input_value)

    # then
    testing_value = logger.level
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_value,expected_value",
    (
        args(
            input="\\tests\\sandbox\\20210106-1715-24124-refinitiv-data-platform-lib.artem.9",
            expected="\\tests\\sandbox\\20210106-1715-9-24124-refinitiv-data-platform-lib.artem",
        ),
        args(
            input="\\tests\\sandbox\\20210106-1715-24124-refinitiv.data.platform.lib.9",
            expected="\\tests\\sandbox\\20210106-1715-9-24124-refinitiv.data.platform.lib",
        ),
        args(
            input="\\tests\\sandbox\\20210106-1715-24124-refinitiv-data-platform-lib.9",
            expected="\\tests\\sandbox\\20210106-1715-9-24124-refinitiv-data-platform-lib",
        ),
        args(
            input="\\tests\\sandbox\\20210106-1715-24124-refinitiv-data-platform-lib..9",
            expected="\\tests\\sandbox\\20210106-1715-9-24124-refinitiv-data-platform-lib.",
        ),
        args(
            input="\\tests\\sandbox\\20210106-1715-24124-refinitiv-data-platform-lib-.9",
            expected="\\tests\\sandbox\\20210106-1715-9-24124-refinitiv-data-platform-lib-",
        ),
        args(
            input="\\tests\\sandbox\\20210106-1715-24124-refinitiv-data-platform-lib.",
            expected="\\tests\\sandbox\\20210106-1715--24124-refinitiv-data-platform-lib",
        ),
        args(
            input="\\tests\\sandbox\\20210106-1715-24124-refinitiv-data-platform-lib-",
            expected="\\tests\\sandbox\\20210106-1715--24124-refinitiv-data-platform-lib-",
        ),
        args(
            input="\\tests\\sandbox\\20210106-1715-24124-refinitiv-data-platform-lib.-",
            expected="\\tests\\sandbox\\20210106-1715--24124-refinitiv-data-platform-lib.-",
        ),
        args(
            input="\\tests\\sandbox\\20210106-1715-24124-refinitiv-data-platform-lib-.",
            expected="\\tests\\sandbox\\20210106-1715--24124-refinitiv-data-platform-lib-",
        ),
    ),
)
def test_filenamer(input_value, expected_value):
    # when
    testing_value = log._filenamer(input_value)

    # then
    assert testing_value == expected_value, testing_value


sep = os.path.sep


@pytest.mark.parametrize(
    "input_value,expected_value",
    (
        args(
            input="\\refinitiv-data-lib.log",
            expected=f"{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input=".\\refinitiv-data-lib.log",
            expected="datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="..\\refinitiv-data-lib.log",
            expected=f"..{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="..\\..\\refinitiv-data-lib.log",
            expected=f"..{sep}..{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="\\..\\refinitiv-data-lib.log",
            expected=f"{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="path\\to\\refinitiv-data-lib.log",
            expected=f"path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input=".\\path\\to\\refinitiv-data-lib.log",
            expected=f"path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="..\\path\\to\\refinitiv-data-lib.log",
            expected=f"..{sep}path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="..\\..\\path\\to\\refinitiv-data-lib.log",
            expected=f"..{sep}..{sep}path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="\\..\\path\\to\\refinitiv-data-lib.log",
            expected=f"{sep}path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="C:\\path\\to\\refinitiv-data-lib.log",
            expected=f"C:{sep}path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="/refinitiv-data-lib.log",
            expected=f"{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="./refinitiv-data-lib.log",
            expected="datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="../refinitiv-data-lib.log",
            expected=f"..{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="../../refinitiv-data-lib.log",
            expected=f"..{sep}..{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="/../refinitiv-data-lib.log",
            expected=f"{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="path/to/refinitiv-data-lib.log",
            expected=f"path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="./path/to/refinitiv-data-lib.log",
            expected=f"path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="../path/to/refinitiv-data-lib.log",
            expected=f"..{sep}path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="../../path/to/refinitiv-data-lib.log",
            expected=f"..{sep}..{sep}path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="/../path/to/refinitiv-data-lib.log",
            expected=f"{sep}path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
        args(
            input="C:/path/to/refinitiv-data-lib.log",
            expected=f"C:{sep}path{sep}to{sep}datetime-datetime-pid-refinitiv-data-lib.log",
        ),
    ),
)
def test__get_filename(input_value, expected_value):
    # when
    testing_value = log._get_filename(input_value, stub_datetime, "pid")

    # then
    assert testing_value == expected_value, testing_value


def test_change_location_in_filename_for_log_file(tmp_path):
    import json

    # given
    test_message = "This is the test message."
    config = {
        "logs": {
            "transports": {
                "file": {
                    "enabled": True,
                    "name": f'{tmp_path / "refinitiv-data-lib.log"}',
                }
            }
        }
    }
    configpath = "refinitiv-data.config.json"
    with open(configpath, "w") as f:
        f.write(json.dumps(config))
        f.flush()

    from refinitiv.data import _configure as configure

    configure.reload()
    reload(log)

    # when
    logger = log.get_logger("location.logfile")
    logfilepath = logger.handlers[0].filename
    logger.warning(test_message)

    # then
    with open(logfilepath, "r") as f:
        logline = f.readline()

    assert test_message in logline

    # teardown
    log.dispose_logger(logger)
    os.remove(configpath)



from unittest.mock import Mock

from refinitiv.data.content._universe_streams import UniverseStreamFacade


def test_stream_available_methods():
    # given
    expected_methods = [
        "close",
        "open",
        "open_async",
        "open_state",
    ]

    # when
    stream = UniverseStreamFacade(stream=Mock())

    # then
    for expected_method in expected_methods:
        assert hasattr(stream, expected_method), expected_method


def test_stream_available_attribute():
    # given
    expected_attribute = ["_stream"]

    # when
    stream = UniverseStreamFacade(...)
    attributes = list(stream.__dict__.keys())

    # then
    assert attributes == expected_attribute



from typing import Optional, TYPE_CHECKING

from ._bid_ask_fields_formula_description import BidAskFieldsFormulaDescription
from ..._enums import InstrumentType
from ...._object_definition import ObjectDefinition


if TYPE_CHECKING:
    from ......_types import OptStr


class FormulaParameterDescription(ObjectDefinition):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    instrument_type : InstrumentType, optional

    instrument_code : str, optional

    fields : BidAskFieldsFormulaDescription, optional

    name : str, optional

    """

    def __init__(
        self,
        instrument_type: Optional[InstrumentType] = None,
        instrument_code: "OptStr" = None,
        fields: Optional[BidAskFieldsFormulaDescription] = None,
        name: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.instrument_type = instrument_type
        self.instrument_code = instrument_code
        self.fields = fields
        self.name = name

    @property
    def fields(self):
        """
        :return: object BidAskFieldsFormulaDescription
        """
        return self._get_object_parameter(BidAskFieldsFormulaDescription, "fields")

    @fields.setter
    def fields(self, value):
        self._set_object_parameter(BidAskFieldsFormulaDescription, "fields", value)

    @property
    def instrument_type(self):
        """
        :return: enum InstrumentType
        """
        return self._get_enum_parameter(InstrumentType, "instrumentType")

    @instrument_type.setter
    def instrument_type(self, value):
        self._set_enum_parameter(InstrumentType, "instrumentType", value)

    @property
    def instrument_code(self):
        """
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)

    @property
    def name(self):
        """
        :return: str
        """
        return self._get_parameter("name")

    @name.setter
    def name(self, value):
        self._set_parameter("name", value)



import numpy as np
import pandas as pd
import pytest

from refinitiv.data.content.ipa._content_provider import (
    CurvesAndSurfacesRequestFactory,
    value_arg_parser,
    parse_value,
    CrossCurrencyCurvesDefinitionsRequestFactory,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions._delete import (
    DeleteRequest,
)
from refinitiv.data.delivery._data._endpoint_data import RequestMethod
from .conftest import StubInstrument, StubOutputs
from ...conftest import StubDefinition


def tests_curves_and_surfaces_request_factory_get_request_method_return_post():
    # given
    expected_value = RequestMethod.POST
    factory = CurvesAndSurfacesRequestFactory()

    # when
    testing_value = factory.get_request_method()

    # then
    assert testing_value == expected_value


def tests_curves_and_surfaces_request_factory_get_body_parameters():
    # given
    factory = CurvesAndSurfacesRequestFactory()

    # when
    testing_value = factory.get_body_parameters(universe=StubDefinition())

    # then
    assert testing_value["universe"] is not None


def tests_cross_currency_curves_definitions_request_factory_get_body_parameters():
    # given
    factory = CrossCurrencyCurvesDefinitionsRequestFactory()

    # when
    testing_value = factory.get_body_parameters(request_items=DeleteRequest(id="42"))

    # then
    assert testing_value == {"id": "42"}


@pytest.mark.parametrize(
    "input_arg",
    [
        None,
        [],
        1,
        0,
        "",
        "dd",
        {},
        [4, 5],
        {"id": 45},
    ],
)
def tests_cross_currency_curves_definitions_request_factory_get_body_parameters_not_valid_arg(
    input_arg,
):
    # given
    factory = CrossCurrencyCurvesDefinitionsRequestFactory()

    # when
    testing_value = factory.get_body_parameters(request_items=input_arg)

    # then
    assert testing_value == {}


def tests_cross_currency_curves_definitions_request_factory_extend_body_parameters():
    # given
    factory = CrossCurrencyCurvesDefinitionsRequestFactory()

    # when
    testing_value = factory.extend_body_parameters(
        body_parameters={"id": 10}, extended_params={"id": 42}
    )

    # then
    assert testing_value == {"id": 42}


def tests_curves_and_surfaces_request_factory_get_body_parameters_pass_instrument():
    # given
    instrument = StubInstrument()
    factory = CurvesAndSurfacesRequestFactory()

    # when
    testing_value = factory.get_body_parameters(universe=instrument)

    # then
    assert testing_value["universe"] is not None


def tests_curves_and_surfaces_request_factory_get_body_parameters_pass_instrument_with_request_item():
    # given
    instrument = StubInstrument()
    instrument._request_item = StubInstrument()
    factory = CurvesAndSurfacesRequestFactory()

    # when
    testing_value = factory.get_body_parameters(universe=instrument)

    # then
    assert testing_value["universe"] is not None


def tests_curves_and_surfaces_request_factory_get_body_parameters_pass_outputs():
    # given
    factory = CurvesAndSurfacesRequestFactory()

    # when
    testing_value = factory.get_body_parameters(
        outputs="outputs", universe=StubDefinition()
    )

    # then
    assert testing_value["outputs"] is not None


def tests_curves_and_surfaces_request_factory_get_body_parameters_pass_outputs_as_enum():
    # given
    factory = CurvesAndSurfacesRequestFactory()

    # when
    testing_value = factory.get_body_parameters(
        outputs=StubOutputs.OUTPUTS, universe=StubDefinition()
    )

    # then
    assert testing_value["outputs"] is not None


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (1, 1),
        ("1", 1),
        (1629099184, 1629099184),
        ("1629099184", 1629099184),
        ("1.01", 1.01),
        (1.01, 1.01),
        ("2021-08-20", np.datetime64("2021-08-20")),
        (np.datetime64("2021-08-20"), np.datetime64("2021-08-20")),
    ],
)
def test_value_arg_parser_successfully(input_value, expected_value):
    # given
    value = value_arg_parser.parse(input_value)

    # then
    assert value == expected_value


def test_value_arg_parser_fail():
    # given
    value = value_arg_parser.parse("")

    # then
    assert pd.isnull(value)


@pytest.mark.parametrize("input_value", [" ", "aaa", "2021-08-20Z147"])
def test_value_arg_parser_try_except(input_value):
    # then
    with pytest.raises(ValueError):
        value_arg_parser.parse(input_value)


@pytest.mark.parametrize(
    ("test_value",),
    [
        ("2018-15-10",),
        ("2018-10-15",),
        ("2018-10-10",),
    ],
)
def test_parse_value(test_value):
    # given
    try:
        # when
        # then
        parse_value(test_value)
    except Exception as e:
        pytest.fail("Unexpected exception", str(e))



def test_public_get_config():
    # given
    import refinitiv.data as rd

    # when
    config = rd.get_config()

    # then
    assert config


def test_reload_method_is_creating_new_config_object():
    # given
    from refinitiv.data import _configure as configure

    prev_config = configure.get_config()

    # when
    configure.reload()
    cur_config = configure.get_config()

    # then
    assert prev_config is not cur_config


def test_no_new_config_object_without_calling_reload_method():
    # given
    from refinitiv.data import _configure as configure

    prev_config = configure.get_config()

    # when
    cur_config = configure.get_config()

    # then
    assert prev_config is cur_config



# coding: utf8

__all__ = ["VolatilitySurfacePoint"]

from ..._object_definition import ObjectDefinition


class VolatilitySurfacePoint(ObjectDefinition):
    def __init__(self, x=None, y=None):
        super().__init__()
        self.x = x
        self.y = y

    @property
    def x(self):
        """
        The coordinate of the volatility data point on the x-axis
        :return: str
        """
        return self._get_parameter("x")

    @x.setter
    def x(self, value):
        self._set_parameter("x", value)

    @property
    def y(self):
        """
        The coordinate of the volatility data point on the y-axis
        :return: str
        """
        return self._get_parameter("y")

    @y.setter
    def y(self, value):
        self._set_parameter("y", value)



import numpy as np
from typing import List, Optional, Union

from ..._types import OptDateTime
from ...content.ipa._enums import DateScheduleFrequency, DayOfWeek
from ...content.ipa.dates_and_calendars.date_schedule import Definition


def date_schedule(
    frequency: Union[DateScheduleFrequency, str] = None,
    start_date: "OptDateTime" = None,
    end_date: "OptDateTime" = None,
    calendar_day_of_month: Optional[int] = None,
    calendars: Optional[List[str]] = None,
    currencies: Optional[List[str]] = None,
    day_of_week: Optional[Union[DayOfWeek, str]] = None,
    count: Optional[int] = None,
) -> List[np.datetime64]:
    """
    With this function you can get date schedules.

    Parameters
    ----------
        frequency: DateScheduleFrequency or str, optional
            The frequency of dates in the predefined period.
        start_date: str or datetime or timedelta, optional
            The start date of the predetermined list of dates.
            The start date must be earlier or equal to the end date.
            Mandatory if endDate is in the past.
        end_date: str or datetime or timedelta, optional
            The end date of the predetermined list of dates.
            If start_date is not set end_date is used to
            define a list of dates from today to the end date;
            end_date and count should not be set at a time;
            end_date must be later or equal to start_date.
            Mandatory if count is not specified.
        calendar_day_of_month : int, optional
            The number of the day of the month to which dates are adjusted.
            The first date in the list is defined as the corresponding
            day of the month to which the start date belongs.
            Mandatory if frequency is set to 'Monthly'.
        calendars: list of str, optional
            Calendars to use the date for working day or weekend.
            Optional if currencies is provided.
        currencies: list of str, optional
            Currencies to use the date for working day or weekend.
            Optional if calendars is provided.
        day_of_week : DayOfWeek or str, optional
            The day of week to which dates are adjusted.
            The first date in the list is defined as
            corresponding day of week following the start date.
            The last date in the list is defined as
            corresponding day of week preceding the end date.
        count : int, optional
            A number is used to define a list of dates from the
            start date (or today if the start day is not set) to the end date.
            Mandatory if end_date is not specified.

    Returns
    -------
    List[np.datetime64]
        List of np.datetime64 dates.

    Examples
    --------
     >>> import datetime
     >>> import refinitiv.data as rd
     >>> from refinitiv.data import dates_and_calendars
     >>>
     >>> rd.open_session("platform.default")
     >>>
     >>> dates = rd.dates_and_calendars.date_schedule(
     ...    start_date=datetime.date(2019, 4, 30),
     ...    count=10,
     ...    frequency=dates_and_calendars.DateScheduleFrequency.WEEKLY,
     ...    calendars=["EMU", "GER"],
     ...    day_of_week="Tuesday",
     >>>)
    """

    response = Definition(
        frequency=frequency,
        start_date=start_date,
        end_date=end_date,
        calendar_day_of_month=calendar_day_of_month,
        calendars=calendars,
        currencies=currencies,
        day_of_week=day_of_week,
        count=count,
    ).get_data()

    return response.data.dates



import abc
from itertools import zip_longest
from typing import Generic, TypeVar, TYPE_CHECKING

from ._endpoint_data import Error
from ..._tools import cached_property

if TYPE_CHECKING:
    from ._parsed_data import ParsedData
    from ._response_factory import ABCDataFactory

T = TypeVar("T")


class ABCResponse(abc.ABC):
    is_success: bool


class BaseResponse(Generic[T], ABCResponse):
    def __init__(
        self,
        is_success: bool,
        parsed_data: "ParsedData",
        data_factory: "ABCDataFactory",
        **kwargs,
    ) -> None:
        self.is_success: bool = is_success
        self._parsed_data = parsed_data
        self._data_factory = data_factory
        self._kwargs = kwargs

        self._status = parsed_data.status
        self.errors = [
            Error(code, msg)
            for code, msg in zip_longest(
                parsed_data.error_codes, parsed_data.error_messages
            )
        ]
        self._raw_response = parsed_data.raw_response
        self.http_response = self._raw_response

    @cached_property
    def data(self):
        return self._data_factory.create_data(
            self._parsed_data, owner_=self, **self._kwargs
        )

    @cached_property
    def requests_count(self):
        if isinstance(self.http_response, list):
            return len(self.http_response)
        return 1

    @property
    def request_message(self):
        if self._raw_response:
            return self._raw_response.request
        return None

    @property
    def closure(self):
        if self._raw_response:
            request = self._raw_response.request
            if isinstance(request, list):
                if isinstance(request[0], list):
                    request = request[0]
                closure = [_request.headers.get("closure") for _request in request]
            else:
                closure = request.headers.get("closure")
            return closure
        return None

    @property
    def http_status(self):
        return self._status

    @property
    def http_headers(self):
        if self._raw_response:
            return self._raw_response.headers
        return None


class Response(BaseResponse):
    pass



from typing import Optional, Union, TYPE_CHECKING

from ._date_schedule_data_provider import DateSchedule
from .._base_request_items import StartEndDateBase
from ...._content_provider_layer import ContentUsageLoggerMixin

from ..._enums import DateScheduleFrequency, DayOfWeek
from ....._content_type import ContentType
from ....._tools import create_repr, try_copy_to_list
from .....delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ....._types import ExtendedParams, OptInt, OptDateTime, OptStrStrs


class DateScheduleRequestItem(StartEndDateBase):
    def __init__(
        self,
        start_date,
        end_date,
        count,
        frequency,
        calendars,
        currencies,
        day_of_week,
        calendar_day_of_month,
    ):
        super().__init__(start_date, end_date)
        self.count = count
        self.frequency = frequency
        self.calendars = calendars
        self.currencies = currencies
        self.day_of_week = day_of_week
        self.calendar_day_of_month = calendar_day_of_month

    @property
    def count(self):
        """
        :return: int
        """
        return self._get_parameter("count")

    @count.setter
    def count(self, value):
        self._set_parameter("count", value)

    @property
    def calendars(self):
        """
        :return: list
        """
        return self._get_parameter("calendars")

    @calendars.setter
    def calendars(self, value):
        self._set_parameter("calendars", value)

    @property
    def currencies(self):
        """
        :return: list
        """
        return self._get_parameter("currencies")

    @currencies.setter
    def currencies(self, value):
        self._set_parameter("currencies", value)

    @property
    def frequency(self):
        """
        :return: DateScheduleFrequency
        """
        return self._get_enum_parameter(DateScheduleFrequency, "frequency")

    @frequency.setter
    def frequency(self, value):
        self._set_enum_parameter(DateScheduleFrequency, "frequency", value)

    @property
    def day_of_week(self):
        """
        :return: DayOfWeek
        """
        return self._get_enum_parameter(DayOfWeek, "DayOfWeek")

    @day_of_week.setter
    def day_of_week(self, value):
        self._set_enum_parameter(DayOfWeek, "DayOfWeek", value)

    @property
    def calendar_day_of_month(self):
        """
        :return: str
        """
        return self._get_parameter("calendarDayOfMonth")

    @calendar_day_of_month.setter
    def calendar_day_of_month(self, value):
        self._set_parameter("calendarDayOfMonth", value)


class Definition(
    ContentUsageLoggerMixin[BaseResponse[DateSchedule]],
    DataProviderLayer[BaseResponse[DateSchedule]],
):
    """
    Date schedule definition object

    Parameters
    ----------
        frequency: DateScheduleFrequency or str
            The frequency of dates in the predefined period.
        start_date: str or datetime or timedelta, optional
            The start date of the predetermined list of dates.
            The start date must be earlier or equal to the end date.
            Mandatory if endDate is in the past.
        end_date: str or datetime or timedelta, optional
            The end date of the predetermined list of dates.
            If start_date is not set end_date is used to
            define a list of dates from today to the end date;
            end_date and count should not be set at a time;
            end_date must be later or equal to start_date.
            Mandatory if count is not specified.
        calendar_day_of_month : int, optional
            The number of the day of the month to which dates are adjusted.
            The first date in the list is defined as the corresponding
            day of the month to which the start date belongs.
            Mandatory if frequency is set to 'Monthly'.
        calendars: list of str, optional
            Calendars to use the date for working day or weekend.
            Optional if currencies is provided.
        currencies: list of str, optional
            Currencies to use the date for working day or weekend.
            Optional if calendars is provided.
        day_of_week : DayOfWeek or str, optional
            The day of week to which dates are adjusted.
            The first date in the list is defined as
            corresponding day of week following the start date.
            The last date in the list is defined as
            corresponding day of week preceding the end date.
        count : int, optional
            A number is used to define a list of dates from the
            start date (or today if the start day is not set) to the end date.
            Mandatory if end_date is not specified.
        extended_params : dict, optional
            If necessary other parameters.

    Methods
    -------
    get_data(session=None, on_response=None, **kwargs)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, **kwargs)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> import datetime
     >>> from refinitiv.data.content.ipa.dates_and_calendars import date_schedule
     >>> definition = date_schedule.Definition(
     ...     start_date="2020-01-01",
     ...     end_date=datetime.timedelta(0),
     ...     frequency="Weekly",
     ...     calendars=["EMU", "GER"],
     ...     day_of_week=date_schedule.DayOfWeek.MONDAY,
     ... )
     >>> response = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    _USAGE_CLS_NAME = "IPA.DatesAndCalendars.DateScheduleDefinition"

    def __init__(
        self,
        frequency: Union[DateScheduleFrequency, str, None] = None,
        start_date: "OptDateTime" = None,
        end_date: "OptDateTime" = None,
        calendar_day_of_month: "OptInt" = None,
        calendars: "OptStrStrs" = None,
        currencies: "OptStrStrs" = None,
        day_of_week: Optional[Union[DayOfWeek, str]] = None,
        count: "OptInt" = None,
        extended_params: "ExtendedParams" = None,
    ):
        calendars = try_copy_to_list(calendars)
        currencies = try_copy_to_list(currencies)
        self.request_item = DateScheduleRequestItem(
            start_date=start_date,
            end_date=end_date,
            count=count,
            frequency=frequency,
            calendars=calendars,
            currencies=currencies,
            calendar_day_of_month=calendar_day_of_month,
            day_of_week=day_of_week,
        )

        super().__init__(
            data_type=ContentType.DATES_AND_CALENDARS_DATE_SCHEDULE,
            universe=self.request_item,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)



from enum import Enum, unique
from typing import Optional, Union

from ..._tools import make_convert_to_enum, EnumArgsParser, make_parse_enum


@unique
class AssetState(Enum):
    """
    Asset state values for 'filter' parameter in request for SymbolConversion content object.
    """

    ACTIVE = "Active"
    INACTIVE = "Inactive"


OptAssetState = Optional[Union[str, AssetState]]

asset_state_enum_arg_parser = EnumArgsParser(
    parse=make_parse_enum(AssetState), parse_to_enum=make_convert_to_enum(AssetState)
)



from time import sleep

import allure
import pytest

from refinitiv.data import open_pricing_stream
from refinitiv.data.discovery import Peers, Screener, Chain
from tests.integration.access.conftest import (
    check_stream_opened,
    check_stream_closed,
    check_stream_snapshot,
    check_stream_universe,
    check_stream_fields,
    check_triggered_events,
)
from tests.integration.helpers import check_the_number_of_items_in_dataframe


@allure.suite("FinCoder layer")
@allure.feature("FinCoder - Open Pricing Stream")
@allure.severity(allure.severity_level.CRITICAL)
class TestOpenPricingStream:
    @allure.title("Open pricing stream - no universe and fields")
    @pytest.mark.caseid("C38695920")
    def test_open_pricing_stream_no_universe_and_fields(self):
        with pytest.raises(TypeError) as error:
            open_pricing_stream()
        assert (
            str(error.value)
            == "open_pricing_stream() missing 1 required positional argument: 'universe'"
        )

    @allure.title("Open pricing stream - with universe and fields")
    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "universe, fields",
        [("GBP=", "BID"), (["USD=", "GBP=", "JPY="], ["BID", "ASK"])],
    )
    @pytest.mark.caseid("C38695923")
    def test_open_pricing_stream_with_universe_and_fields(
        self, open_desktop_session, universe, fields
    ):
        callback_history = []
        stream = open_pricing_stream(
            universe,
            fields,
            on_data=lambda data, ric, stream: callback_history.append(data),
        )
        check_stream_opened(stream)
        sleep(3)
        check_stream_universe(stream, universe)
        check_stream_fields(stream, universe, fields)
        stream.close()
        check_stream_closed(stream)
        check_triggered_events(callback_history, universe)

    @allure.title("Open pricing stream - get snapshot with universe and fields")
    @pytest.mark.parametrize(
        "universe, fields", ((["USD=", "AUD=", "CAD="], ["BID", "ASK", "OPEN_PRC"]),)
    )
    @pytest.mark.caseid("C38695960")
    def test_open_pricing_stream_snapshot_with_universe_and_fields(
        self, open_desktop_session, universe, fields
    ):
        stream = open_pricing_stream(
            universe,
            fields,
        )
        check_stream_opened(stream)
        check_stream_snapshot(stream, universe[:2], fields[:2])

    @allure.title("Open pricing stream - get snapshot with wrong universe and fields")
    @pytest.mark.parametrize(
        "universe, fields, err_msg_universe, err_msg_fields",
        (
            (
                ["USD=", "AUD=", "CAD="],
                ["BID", "ASK", "OPEN_PRC"],
                "Error code -1 | Instrument {'BYN='} was not requested",
                "Error code -1 | Field {'MID_CLOSE'} was not requested",
            ),
        ),
    )
    @pytest.mark.caseid("C38695961")
    def test_open_pricing_stream_snapshot_with_wrong_universe(
        self,
        capsys,
        open_desktop_session,
        universe,
        fields,
        err_msg_universe,
        err_msg_fields,
    ):
        stream = open_pricing_stream(universe, fields)
        check_stream_opened(stream)
        df = stream.get_snapshot("BYN=")
        test_log = capsys.readouterr()
        assert df.empty, f"Snapshot dataframe with wrong universe should be empty"
        assert (
            err_msg_universe in test_log.err
        ), f"Message '{err_msg_universe}' wasn't found in log"

        df = stream.get_snapshot("USD=", ["MID_CLOSE"])
        test_log = capsys.readouterr()
        assert df.empty, f"Snapshot dataframe with wrong fields should be empty"
        assert (
            err_msg_fields in test_log.err
        ), f"Message '{err_msg_fields}' wasn't found in log"

    @allure.title("Open pricing stream - with function and chain as universe")
    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "universe, fields, expected_universe",
        [
            (Peers("VOD.L"), "BID", "CLNX.MC"),
            pytest.param(Chain("0#.DJI"), None, ["IBM.N", "CSCO.OQ"], id="chain"),
            (
                Screener(
                    'U(IN(Equity(active,public,primary))/*UNV:Public*/), IN(TR.HQCountryCode,"AR"), IN(TR.GICSIndustryCode,"401010")'
                ),
                ["BID", "ASK"],
                ["BMA.BA", "BBAR.BA", "BHIP.BA"],
            ),
        ],
    )
    @pytest.mark.caseid("C43874350")
    def test_open_pricing_stream_with_function_and_chain_as_universe(
        self,
        request,
        open_desktop_session,
        universe,
        fields,
        expected_universe,
    ):
        stream = open_pricing_stream(
            universe,
            fields,
        )
        check_stream_opened(stream)
        sleep(1)
        check_stream_universe(stream, expected_universe)
        check_stream_fields(stream, expected_universe, fields)

        if "chain" in request.node.callspec.id:
            df = stream.get_snapshot()
            list_of_instruments = list(df["Instrument"])
            check_the_number_of_items_in_dataframe(
                expected_universe[0], list_of_instruments, 1
            )
        stream.close()
        check_stream_closed(stream)

    @allure.title("Open pricing stream - with CustomInstrument objects")
    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "universe,expected_fields",
        [
            (
                ["EUR=", "GBP="],
                ["BID", "ASK", "TRDPRC_1", "TRADE_DATE", "SALTIM_NS"],
            ),
            (None, ["TRDPRC_1", "TRADE_DATE", "SALTIM_NS"]),
        ],
        ids=["universe_mix", "ci_multi_universe"],
    )
    @pytest.mark.caseid("C43973497")
    def test_open_pricing_stream_with_custom_instrument_objects(
        self,
        open_desktop_session,
        create_instrument,
        universe,
        expected_fields,
        request,
    ):
        fields = []
        symbol_01 = create_instrument()
        symbol_02 = create_instrument()
        if "universe_mix" in request.node.callspec.id:
            universe += [symbol_01, symbol_02]
            fields = expected_fields
        if "ci_multi_universe" in request.node.callspec.id:
            universe = [symbol_01, symbol_02]

        callback_history = []
        stream = open_pricing_stream(
            universe=universe,
            fields=fields,
            on_data=lambda data, ric, stream_item: callback_history.append(data),
        )
        check_stream_opened(stream)
        sleep(2)
        check_stream_universe(stream, universe)
        check_stream_fields(stream, universe, expected_fields)
        stream.close()
        check_stream_closed(stream)
        check_triggered_events(callback_history, universe)



import copy

import pytest

from refinitiv.data.content import historical_pricing
from refinitiv.data.content.historical_pricing import events
from refinitiv.data.errors import RDError
from tests.unit.conftest import StubSession, error_user_has_no_permissions
from tests.unit.content.historical_pricing.data_for_tests import (
    DF_REQUIREMENT_TWO_INSTS_ONE_IS_BAD_WITHOUT_FIELDS,
    DF_WITH_EXTENDED_PARAMS,
)


def test_get_data_when_user_has_no_permissions():
    session = StubSession(is_open=True)
    session.http_responses = error_user_has_no_permissions()
    session.async_mode = True
    with pytest.raises(RDError):
        events.Definition("boo").get_data(session)


@pytest.mark.asyncio
async def test_get_data_async_no_exception_when_user_has_no_permissions():
    try:
        session = StubSession(is_open=True)
        session.http_responses = error_user_has_no_permissions()
        session.async_mode = True
        await events.Definition("boo").get_data_async(session)
    except Exception as e:
        assert False, str(e)
    else:
        assert True


def test_df_requirement_two_insts_one_is_bad_without_fields():
    # given
    session = StubSession(
        is_open=True, response=DF_REQUIREMENT_TWO_INSTS_ONE_IS_BAD_WITHOUT_FIELDS
    )
    expected_str = (
        "                              GBP=                                                                                                   G234BP=                                                                                      \n"
        "                        EVENT_TYPE   RTL     BID     ASK MID_PRICE DSPLY_NAME SRC_REF1 DLG_CODE1      CTBTR_1 CTB_LOC1 QUALIFIERS EVENT_TYPE   RTL   BID   ASK MID_PRICE DSPLY_NAME SRC_REF1 DLG_CODE1 CTBTR_1 CTB_LOC1 QUALIFIERS\n"
        "Timestamp                                                                                                                                                                                                                         \n"
        "2022-06-23 11:54:12.169      quote  8462  1.2219  1.2223    1.2221       <NA>     BCFX      <NA>     BARCLAYS      LON       <NA>       <NA>  <NA>  <NA>  <NA>      <NA>       <NA>     <NA>      <NA>    <NA>     <NA>       <NA>\n"
        "2022-06-23 11:54:12.619      quote  8526   1.222  1.2225   1.22225       <NA>     RAB1      RABX  RABOBANKGFM      LON       <NA>       <NA>  <NA>  <NA>  <NA>      <NA>       <NA>     <NA>      <NA>    <NA>     <NA>       <NA>\n"
        "2022-06-23 11:54:12.757      quote  8590  1.2219  1.2224   1.22215       <NA>     NBJX      NBJJ  NEDBANK LTD      JHB       <NA>       <NA>  <NA>  <NA>  <NA>      <NA>       <NA>     <NA>      <NA>    <NA>     <NA>       <NA>"
    )
    definition = historical_pricing.events.Definition(
        universe=["GBP=", "G234BP="], count=3
    )
    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


def test_df_with_expected_params():
    # given
    session = StubSession(is_open=True, response=DF_WITH_EXTENDED_PARAMS)
    expected_str = (
        "                             VOD.L                                                                                          GOOG.O                                                                                                                                                    \n"
        "                        EVENT_TYPE    RTL   SEQNUM     BID BIDSIZE   ASK  ASKSIZE MID_PRICE IMB_SH IMB_SIDE  QUALIFIERS EVENT_TYPE    RTL     BID BIDSIZE BID_MMID1     ASK ASKSIZE ASK_MMID1 MID_PRICE UPLIMIT LOLIMIT LIMIT_INDQ SH_SAL_RES                               QUALIFIERS\n"
        "Timestamp                                                                                                                                                                                                                                                                             \n"
        "2022-07-19 15:39:01.320      quote  50480  9075524  130.94   80476   131  2453541    130.97   <NA>     <NA>  [ASK_TONE]       <NA>   <NA>    <NA>    <NA>      <NA>    <NA>    <NA>      <NA>      <NA>    <NA>    <NA>       <NA>       <NA>                                     <NA>\n"
        "2022-07-19 15:39:38.390      quote  50496  9075694  130.94   80476   131  2503541    130.97   <NA>     <NA>  [ASK_TONE]       <NA>   <NA>    <NA>    <NA>      <NA>    <NA>    <NA>      <NA>      <NA>    <NA>    <NA>       <NA>       <NA>                                     <NA>\n"
        "2022-07-19 15:40:00.112      quote  50528  9075850  130.94   80476   131  2203541    130.97   <NA>     <NA>  [ASK_TONE]       <NA>   <NA>    <NA>    <NA>      <NA>    <NA>    <NA>      <NA>      <NA>    <NA>    <NA>       <NA>       <NA>                                     <NA>\n"
        "2022-07-19 15:40:07.641      quote  50544  9075948  130.94   80476   131  2153541    130.97   <NA>     <NA>  [ASK_TONE]       <NA>   <NA>    <NA>    <NA>      <NA>    <NA>    <NA>      <NA>      <NA>    <NA>    <NA>       <NA>       <NA>                                     <NA>\n"
        "2022-07-19 15:52:52.900      quote  50624  9076320  130.82   12365   131  2153541    130.91   <NA>     <NA>  [BID_TONE]       <NA>   <NA>    <NA>    <NA>      <NA>    <NA>    <NA>      <NA>      <NA>    <NA>    <NA>       <NA>       <NA>                                     <NA>\n"
        "2022-07-19 15:59:58.409       <NA>   <NA>     <NA>    <NA>    <NA>  <NA>     <NA>      <NA>   <NA>     <NA>        <NA>      quote  60320  113.33     200       DEX  113.35     100       MMX      <NA>  118.44  107.16        BOE          N     [PRC_QL_CD];   [PRC_QL3];A[GV1_TEXT]\n"
        "2022-07-19 15:59:58.409       <NA>   <NA>     <NA>    <NA>    <NA>  <NA>     <NA>      <NA>   <NA>     <NA>        <NA>      quote  60272  113.33     200       DEX  113.36     200       NAS      <NA>  118.44  107.16        BOE          N     [PRC_QL_CD];   [PRC_QL3];A[GV1_TEXT]\n"
        "2022-07-19 15:59:58.410       <NA>   <NA>     <NA>    <NA>    <NA>  <NA>     <NA>      <NA>   <NA>     <NA>        <NA>      quote  60336  113.33     200       DEX  113.35     200       NAS      <NA>  118.44  107.16        BOE          N     [PRC_QL_CD];   [PRC_QL3];A[GV1_TEXT]\n"
        "2022-07-19 15:59:58.424       <NA>   <NA>     <NA>    <NA>    <NA>  <NA>     <NA>      <NA>   <NA>     <NA>        <NA>      quote  60352  113.33     200       DEX  113.35     100       MMX      <NA>  118.44  107.16        BOE          N     [PRC_QL_CD];   [PRC_QL3];A[GV1_TEXT]\n"
        "2022-07-19 15:59:58.425       <NA>   <NA>     <NA>    <NA>    <NA>  <NA>     <NA>      <NA>   <NA>     <NA>        <NA>      quote  60400  113.33     200       DEX  113.35     100       NAS      <NA>  118.44  107.16        BOE          N     [PRC_QL_CD];   [PRC_QL3];A[GV1_TEXT]"
    )
    definition = historical_pricing.events.Definition(
        universe=["VOD.L", "GOOG.O"],
        eventTypes="quote",
        extended_params={
            "count": 5,
            "start": "2022-07-18T10:00:00",
            "end": "2022-07-19T16:00:00",
        },
    )
    # when
    response = definition.get_data(session=session)
    df = response.data.df

    # then
    assert df.to_string() == expected_str


def test_fields_is_not_change():
    expected_fields = ["BID"]
    testing_fields = ["BID"]
    session = StubSession(is_open=True, response=DF_WITH_EXTENDED_PARAMS)
    definition = historical_pricing.events.Definition(
        universe=["GBP=", "G234BP="], count=10, fields=testing_fields
    )
    definition.get_data(session=session)
    assert testing_fields == expected_fields



import datetime
from dataclasses import dataclass
from typing import List, TYPE_CHECKING


import pandas as pd

from ._tools import _get_headline_from_story
from ._udf_html_parser import HeadlineHTMLParser
from ._urgency import Urgency

if TYPE_CHECKING:
    from pandas.core.tools.datetimes import DatetimeScalar


@dataclass
class NewsStoryContent:
    html: "str"
    text: "str"


class NewsData(object):
    def __init__(
        self,
        title: str,
        creator: str,
        source: List[dict],
        language: List[dict],
        item_codes: List[str],
        urgency: int,
        creation_date: str,
        update_date: str,
        raw: dict,
        news_type: str,
        html: str = None,
        text: str = None,
    ) -> None:
        self.title = title
        self.creator = creator
        self.source = source
        self.language = language
        self.item_codes = item_codes
        self.urgency: Urgency = Urgency(urgency)

        if news_type == "story":
            self.content: NewsStoryContent = NewsStoryContent(html, text)
            self.headline: str = _get_headline_from_story(raw)
            self.creation_date: "DatetimeScalar" = pd.to_datetime(creation_date)
            self.update_date: "DatetimeScalar" = pd.to_datetime(update_date)
        elif news_type == "headline":
            self.first_created: "DatetimeScalar" = pd.to_datetime(creation_date)
            self.version_created: "DatetimeScalar" = pd.to_datetime(update_date)
            self.story_id: str = raw["storyId"]


class HeadlineRDP(NewsData):
    @staticmethod
    def from_dict(datum: dict) -> "HeadlineRDP":
        news_item = datum.get("newsItem")
        content_meta = news_item.get("contentMeta")
        item_meta = news_item.get("itemMeta")
        subject = content_meta.get("subject")
        headline = HeadlineRDP(
            title=item_meta.get("title")[0].get("$"),
            creator=content_meta.get("creator")[0].get("_qcode"),
            source=content_meta.get("infoSource"),
            language=content_meta.get("language"),
            item_codes=[item.get("_qcode") for item in subject],
            urgency=content_meta.get("urgency").get("$"),
            creation_date=item_meta.get("firstCreated").get("$"),
            update_date=item_meta.get("versionCreated").get("$"),
            raw=datum,
            news_type="headline",
        )
        return headline


class HeadlineUDF:
    def __init__(
        self,
        display_direction,
        document_type,
        first_created,
        is_alert,
        language,
        report_code,
        source_name,
        story_id,
        text,
        version_created,
    ) -> None:
        self.display_direction = display_direction
        self.document_type = document_type
        self.first_created = pd.to_datetime(first_created)
        self.is_alert = is_alert
        self.language = language
        self.report_code = report_code
        self.source_name = source_name
        self.story_id = story_id
        self.text = text
        self.version_created = pd.to_datetime(version_created)

    @classmethod
    def from_dict(cls, datum: dict) -> "HeadlineUDF":
        display_direction = datum.get("displayDirection")
        document_type = datum.get("documentType")
        first_created = datum.get("firstCreated")
        is_alert = datum.get("isAlert")
        language = datum.get("language")
        report_code = datum.get("reportCode")
        source_name = datum.get("sourceName")
        story_id = datum.get("storyId")
        text = datum.get("text")
        version_created = datum.get("versionCreated")
        headline = cls(
            display_direction=display_direction,
            document_type=document_type,
            first_created=first_created,
            is_alert=is_alert,
            language=language,
            report_code=report_code,
            source_name=source_name,
            story_id=story_id,
            text=text,
            version_created=version_created,
        )
        return headline


class Story(NewsData):
    @classmethod
    def from_dict(cls, datum: dict) -> "Story":
        news_item = datum.get("newsItem")
        content_meta = news_item.get("contentMeta")
        html = news_item.get("contentSet", {}).get("inlineXML", [{}])[0].get("$")
        text = news_item.get("contentSet", {}).get("inlineData", [{}])[0].get("$")
        item_meta = news_item.get("itemMeta")
        subject = content_meta.get("subject")
        story = cls(
            title=item_meta.get("title")[0].get("$"),
            creator=content_meta.get("creator")[0].get("_qcode"),
            source=content_meta.get("infoSource"),
            language=content_meta.get("language"),
            item_codes=[item.get("_qcode") for item in subject],
            urgency=content_meta.get("urgency").get("$"),
            creation_date=item_meta.get("firstCreated").get("$"),
            update_date=item_meta.get("versionCreated").get("$"),
            raw=datum,
            news_type="story",
            html=html,
            text=text,
        )
        return story


class StoryUDF:
    def __init__(
        self,
        title: str,
        text: str,
        html: str,
        creation_date: datetime.datetime,
        headline: str,
        creator=None,
    ) -> None:
        self.title = title
        self.content: NewsStoryContent = NewsStoryContent(html, text)
        self.creator = creator
        self.creation_date = creation_date
        self.headline = headline

    @classmethod
    def from_dict(cls, datum: dict) -> "StoryUDF":
        parser = HeadlineHTMLParser()
        story = datum["story"]
        for i in story:
            parser.feed(story[i])
        text = parser.data.get("text") or []
        text = "\n".join(text)
        html = parser.data.get("html") or datum.get("story", {}).get("storyHtml") or []
        story = cls(
            title=parser.data.get("headline"),
            text=text,
            html=html,
            creator=parser.data.get("creator"),
            creation_date=pd.to_datetime(parser.data.get("creation_date")),
            headline=parser.data.get("headline"),
        )
        return story



import inspect
from unittest import mock

import pandas as pd
import pytest

from refinitiv.data.content import pricing
from refinitiv.data.content.pricing._pricing_content_provider import pricing_build_df
from refinitiv.data.errors import RDError
from tests.unit.conftest import StubConfig, StubResponse
from tests.unit.conftest import StubSession
from tests.unit.content.data_for_tests import PRICING_DF_CREATING_RESPONSE


def test_call_get_data_without_session():
    # given
    definition = pricing.Definition("universe")

    # then
    with pytest.raises(Exception):
        # when
        definition.get_data()


def test_call_get_data_with_session():
    # given
    session = StubSession(
        is_open=True,
        config=StubConfig(
            {
                "apis.data.pricing": {
                    "url": "test",
                    "endpoints.snapshots": "test",
                },
                "raise_exception_on_error": True,
            }
        ),
        response=StubResponse(content_data=[{"Type": {}, "Fields": {}}]),
    )
    definition = pricing.Definition("universe")

    # when
    testing_value = definition.get_data(session)

    # then
    assert testing_value


def test_get_stream():
    # given
    session = StubSession(is_open=True)
    definition = pricing.Definition("universe")

    # when
    stream = definition.get_stream(session)
    stream._stream._log

    # then
    assert stream


def test_get_fields_with_fields():
    # given
    expected_value = {"EUR=": {"field1": "aaa"}}
    session = StubSession(is_open=True)
    definition = pricing.Definition("universe")
    stream = definition.get_stream(session=session)

    with mock.patch(
        "refinitiv.data.content._universe_streams._UniverseStreams.__getitem__",
        return_value={"field1": "aaa", "field2": "bbb"},
    ):
        # when
        testing_value = stream._get_fields("EUR=", ["field1"])

        # then
        assert testing_value == expected_value


def test_get_fields_without_fields():
    # given
    expected_value = {"EUR=": {"field1": "aaa", "field2": "bbb"}}
    definition = pricing.Definition("universe")
    session = StubSession(is_open=True)
    stream = definition.get_stream(session=session)

    with mock.patch(
        "refinitiv.data.content._universe_streams._UniverseStreams.__getitem__",
        return_value={"field1": "aaa", "field2": "bbb"},
    ):
        # when
        testing_value = stream._get_fields("EUR=")

        # then
        assert testing_value == expected_value


def test_open_stream_with_empty_list_as_universe():
    session = StubSession(is_open=True)

    stream = pricing.Definition(universe=[], fields=["BID", "ASK"]).get_stream(session)

    with pytest.raises(ValueError):
        stream.open()


def test_pricing_definition__repr__():
    # given
    definition = pricing.Definition(universe="universe")
    hex_id = hex(id(definition))
    expected_value = (
        f"<refinitiv.data.content.pricing.Definition object at {hex_id} "
        f"{{name=['universe']}}>"
    )

    # when
    testing_value = repr(definition)

    # then
    assert testing_value == expected_value


def test_input_user_attributes():
    # given
    excepted_attributes = ["_universe", "_fields", "_service", "_extended_params"]
    definition = pricing.Definition("")

    # when
    attributes = definition.__dict__

    # then
    assert all(x in attributes for x in excepted_attributes)


def test_inspect_init_pricing():
    # given
    excepted_attributes = [
        "universe",
        "fields",
        "service",
        "extended_params",
    ]
    inspect_pricing_init = inspect.signature(pricing.Definition.__init__)

    # when
    attributes = inspect_pricing_init.parameters

    # then
    assert all(x in attributes for x in excepted_attributes)


def test_smoke_get_stream_into_pricing():
    # given
    session = StubSession(is_open=True)
    definition = pricing.Definition("universe")

    with mock.patch(
        "refinitiv.data.content.pricing._stream_facade.Stream.__init__",
        return_value=None,
    ) as mock_stream:
        # when
        definition.get_stream(session=session)

        # then
        mock_stream.assert_called_once_with(
            universe=["universe"],
            session=session,
            fields=[],
            service=None,
            api=None,
            extended_params={},
        )


def test_error_message_for_more_100_rics():
    # given
    response = StubResponse(
        content_data={
            "error": {
                "id": "6c44f198-b44e-42cb-b7a9-ce0ddb792256",
                "code": "400",
                "message": "Validation error",
                "status": "Bad Request",
                "errors": [
                    {
                        "key": "universe",
                        "reason": "validation failure list:\nuniverse in query should "
                        "have at most 100 instruments",
                    }
                ],
            }
        },
        status_code=400,
    )
    session = StubSession(is_open=True, response=response)
    definition = pricing.Definition(universe=["100Rics"], fields=["BID", "ASK"])

    # then
    error_message = (
        "Error code 400 | Validation error. validation failure "
        "list:\nuniverse in query should have at most 100 instruments"
    )
    with pytest.raises(RDError, match=error_message):
        # when
        definition.get_data(session)


@pytest.mark.parametrize("convert", [True, False])
def test_pricing_get_snapshot_should_not_return_none_in_df(convert):
    # given
    from refinitiv.data.content._universe_streams import build_df

    universe = ["AIR.PA", "CAPP.PA"]
    fields = ["PDTRDTM_MS", "VEH_PERMID"]
    values_by_field = {"PDTRDTM_MS": [None, None], "VEH_PERMID": [None, None]}

    # when
    df = build_df(universe, fields, values_by_field, convert)

    # then
    """
      Instrument PDTRDTM_MS VEH_PERMID
    0     AIR.PA       <NA>       <NA>
    1    CAPP.PA       <NA>       <NA>
    """
    assert df["VEH_PERMID"][0] is pd.NA
    assert df["PDTRDTM_MS"][1] is pd.NA


@pytest.mark.parametrize("convert", [True, False])
def test_pricing_get_snapshot_should_not_return_none_in_df_another_case(convert):
    # given
    from refinitiv.data.content._universe_streams import build_df

    universe = ["EUP=", "GBP=", "JPY=", "CAD="]
    fields = ["BID", "ASK"]
    values_by_field = {
        "BID": [None, 1.2085, 134.49, 1.2937],
        "ASK": [None, 1.2082, 134.46, 1.2932],
    }

    # when
    df = build_df(universe, fields, values_by_field, convert)

    # then
    expected_str = (
        "  Instrument     BID     ASK\n"
        "0       EUP=    <NA>    <NA>\n"
        "1       GBP=  1.2085  1.2082\n"
        "2       JPY=  134.49  134.46\n"
        "3       CAD=  1.2937  1.2932"
    )
    assert df.to_string() == expected_str


def test_universe_and_fields_order_get_snapshot_without_params():
    # given
    expected_snapshot = (
        "  Instrument BField AField DField CField\n"
        "0      BInst   <NA>   <NA>   <NA>   <NA>\n"
        "1      AInst   <NA>   <NA>   <NA>   <NA>\n"
        "2      DInst   <NA>   <NA>   <NA>   <NA>\n"
        "3      CInst   <NA>   <NA>   <NA>   <NA>"
    )
    session = StubSession(is_open=True)
    definition = pricing.Definition(
        universe=["BInst", "AInst", "DInst", "CInst"],
        fields=["BField", "AField", "DField", "CField"],
    )
    stream = definition.get_stream(session=session)

    # when
    snapshot = stream.get_snapshot()

    # then
    assert snapshot.to_string() == expected_snapshot


def test_universe_and_fields_order_get_snapshot_with_universe():
    # given
    expected_snapshot = (
        "  Instrument BField AField DField CField\n"
        "0      DInst   <NA>   <NA>   <NA>   <NA>\n"
        "1      AInst   <NA>   <NA>   <NA>   <NA>\n"
        "2      BInst   <NA>   <NA>   <NA>   <NA>"
    )
    session = StubSession(is_open=True)
    definition = pricing.Definition(
        universe=["BInst", "AInst", "DInst", "CInst"],
        fields=["BField", "AField", "DField", "CField"],
    )
    stream = definition.get_stream(session=session)

    # when
    snapshot = stream.get_snapshot(universe=["DInst", "AInst", "BInst"])

    # then
    assert snapshot.to_string() == expected_snapshot


def test_universe_and_fields_order_get_snapshot_with_fields():
    # given
    expected_snapshot = (
        "  Instrument DField AField BField\n"
        "0      BInst   <NA>   <NA>   <NA>\n"
        "1      AInst   <NA>   <NA>   <NA>\n"
        "2      DInst   <NA>   <NA>   <NA>\n"
        "3      CInst   <NA>   <NA>   <NA>"
    )
    session = StubSession(is_open=True)
    definition = pricing.Definition(
        universe=["BInst", "AInst", "DInst", "CInst"],
        fields=["BField", "AField", "DField", "CField"],
    )
    stream = definition.get_stream(session=session)

    # when
    snapshot = stream.get_snapshot(fields=["DField", "AField", "BField"])

    # then
    assert snapshot.to_string() == expected_snapshot


def test_return_none_if_no_field_not_raise_error():
    # given
    session = StubSession(is_open=True)
    definition = pricing.Definition(
        ["EUP=", "GBP=", "JPY=", "CAD="],
        fields=["BID", "ASK"],
    )
    non_streaming = definition.get_stream(session=session)

    # then
    try:
        # when
        testing_value = non_streaming["EUP="]["TEST"]
    except Exception as e:
        assert False, str(e)

    assert testing_value is None


def test_return_none_if_no_universe_not_raise_error():
    # given
    session = StubSession(is_open=True)
    definition = pricing.Definition(
        ["EUP=", "GBP=", "JPY=", "CAD="],
        fields=["BID", "ASK"],
    )
    non_streaming = definition.get_stream(session=session)

    # then
    try:
        # when
        testing_value = non_streaming["TEST"]
    except Exception as e:
        assert False, str(e)

    assert testing_value == {}


def test_get_snapshot_can_expect_one_string_in_fields():
    # given
    session = StubSession(is_open=True)
    definition = pricing.Definition(
        ["EUP=", "GBP=", "JPY=", "CAD="],
        fields=["BID", "ASK"],
    )
    non_streaming = definition.get_stream(session=session)

    # when
    testing_value = non_streaming.get_snapshot("EUP=", fields="BID")

    # then
    try:
        testing_value["BID"]
    except KeyError as e:
        assert False, str(e)
    else:
        assert True


@pytest.mark.parametrize(
    "input_fields, expected_fields",
    [
        (["BID", "ASK", "BID", "ASK"], ["BID", "ASK"]),
        # order is important
        (["EUP=", "GBP=", "JPY=", "CAD="], ["EUP=", "GBP=", "JPY=", "CAD="]),
        ("EUP=", ["EUP="]),
        ([], []),
        (None, []),
        ("", []),
    ],
)
def test_pricing_definition_can_expect_fields(input_fields, expected_fields):
    # when
    definition = pricing.Definition("EUP=", fields=input_fields)
    testing_fields = definition._fields

    # then
    assert testing_fields == expected_fields


def test_pricing_build_df_none_to_na():
    raw = [{"Type": "Refresh", "Fields": {"NONE_FIELD": None}}]
    universe = ["EUR="]

    result = pricing_build_df(raw, universe, [])

    # assert that universe is the first dataframe column
    assert result.at[0, "Instrument"] == "EUR="

    # assert that None turned into pd.NA in result dataframe
    assert result.at[0, "NONE_FIELD"] is pd.NA


def test_pricing_build_df_item_type_status():
    raw = [
        {
            "Type": "Status",
            "State": {"Code": "NotFound"},
        }
    ]
    universe = ["USD="]
    fields = ["FIRST_FIELD", "SECOND_FIELD"]
    result = pricing_build_df(raw, universe, fields)

    assert result.at[0, "Instrument"] == "USD="
    assert len(result.columns) == 3  # RIC and two fields
    assert result.at[0, "FIRST_FIELD"] == "#N/F"
    assert result.at[0, "SECOND_FIELD"] == "#N/F"


def test_df_without_nan():
    expected_str = (
        "  Instrument CF_NAME  DDS_DSO_ID  BR_LINK1  QUOTIM_2  QUOTE_DT2     ASKHI1_MS\n"
        "0       EUR=    Euro       12348      <NA>  10:30:40 2022-06-22  22:55:25.876"
    )
    session = StubSession(is_open=True, response=PRICING_DF_CREATING_RESPONSE)
    definition = pricing.Definition("EUR=")
    response = definition.get_data(session=session)
    df = response.data.df
    assert df.to_string() == expected_str


def test_add_instruments():
    # given
    test_universe = ["VOD.L", "IBM.N"]
    test_instruments = ["EUR="]
    test_result = ["VOD.L", "IBM.N", "EUR="]
    session = StubSession(is_open=True)
    stream = pricing.Definition(universe=test_universe).get_stream(session=session)

    # when
    stream.add_instruments(test_instruments)

    # then
    assert stream._universe == test_result


def test_remove_instruments():
    # given
    test_universe = ["VOD.L", "IBM.N"]
    test_instruments = ["IBM.N"]
    test_result = ["VOD.L"]
    session = StubSession(is_open=True)
    stream = pricing.Definition(universe=test_universe).get_stream(session=session)

    # when
    stream.remove_instruments(test_instruments)

    # then
    assert stream._universe == test_result


def test_add_fields():
    # given
    test_fields = ["BID", "ASK"]
    test_new_fields = ["TRDPRC_1"]
    test_result = ["BID", "ASK", "TRDPRC_1"]
    session = StubSession(is_open=True)
    stream = pricing.Definition(universe=["VOD.L"], fields=test_fields).get_stream(
        session=session
    )

    # when
    stream.add_fields(test_new_fields)

    # then
    assert stream._fields == test_result


def test_remove_fields():
    # given
    test_fields = ["BID", "ASK"]
    test_remove_fields = ["ASK"]
    test_result = ["BID"]
    session = StubSession(is_open=True)
    stream = pricing.Definition(universe=["VOD.L"], fields=test_fields).get_stream(
        session=session
    )

    # when
    stream.remove_fields(test_remove_fields)

    # then
    assert stream._fields == test_result



from unittest.mock import patch

import mock
import pytest

from refinitiv.data._core.session import set_default
from refinitiv.data._errors import RDError, ScopeError
from refinitiv.data.discovery import Peers, Screener, Chain
from tests.unit.conftest import StubSession, StubResponse, send_ws_messages
from tests.unit.delivery.stream.conftest import StubWebSocketApp
from .data_for_tests import (
    BUNCH_CONSTITUENTS,
    EXPECTED_CONSTITUENTS_LIST,
    EXPECTED_SUMMARY_LIST,
)

universe = ["BT.L", "DTEGn.DE", "TEF.MC", "ORAN.PA"]
valid_adc_response = StubResponse(
    {
        "responses": [
            {
                "columnHeadersCount": 1,
                "data": [
                    ["BT.L", "BT.L"],
                    ["DTEGn.DE", "DTEGn.DE"],
                    ["TEF.MC", "TEF.MC"],
                    ["ORAN.PA", "ORAN.PA"],
                ],
                "headerOrientation": "horizontal",
                "headers": [
                    [
                        {"displayName": "Instrument"},
                        {"displayName": "RIC", "field": "TR.RIC"},
                    ]
                ],
                "rowHeadersCount": 1,
                "totalColumnsCount": 2,
                "totalRowsCount": 4,
            }
        ]
    }
)
adc_response_with_error_screen = StubResponse(
    {
        "responses": [
            {
                "columnHeadersCount": 1,
                "data": [["screen(IBM.NNN)", None]],
                "error": [
                    {
                        "code": 800,
                        "col": 1,
                        "message": "SCREEN(IBM.NNN) processing failed.",
                        "row": 0,
                    }
                ],
                "headerOrientation": "horizontal",
                "headers": [
                    [
                        {"displayName": None},
                        {"displayName": "TR.RIC", "field": "TR.RIC"},
                    ]
                ],
                "rowHeadersCount": 1,
                "totalColumnsCount": 2,
                "totalRowsCount": 2,
            }
        ]
    }
)

infos = {
    "services": [
        {
            "port": 443,
            "location": ["ap-northeast-1a"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "ap-northeast-1-aws-1-sm.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
        },
    ]
}

stream_data = [
    {
        "ID": 5,
        "Type": "Refresh",
        "Key": {"Service": "ELEKTRON_DD", "Name": "0#.DJI"},
        "State": {"Stream": "Open", "Data": "Ok"},
        "Qos": {"Timeliness": "Realtime", "Rate": "JitConflated"},
        "PermData": "AwEBlWnA",
        "SeqNumber": 3216,
        "Fields": {
            "PROD_PERM": 9569,
            "RDNDISPLAY": 187,
            "DSPLY_NAME": "DJ INDU AVERAGE",
            "RDN_EXCHID": "   ",
            "CURRENCY": "USD",
            "REF_COUNT": 14,
            "RECORDTYPE": 120,
            "LONGLINK1": ".DJI",
            "LONGLINK2": "AAPL.OQ",
            "LONGLINK3": "AMGN.OQ",
            "LONGLINK4": "AXP.N",
            "LONGLINK5": "BA.N",
            "LONGLINK6": "CAT.N",
            "LONGLINK7": "CRM.N",
            "LONGLINK8": "CSCO.OQ",
            "LONGLINK9": "CVX.N",
            "LONGLINK10": "DIS.N",
            "LONGLINK11": "DOW.N",
            "LONGLINK12": "GS.N",
            "LONGLINK13": "HD.N",
            "LONGLINK14": "HON.OQ",
            "LONGPREVLR": None,
            "LONGNEXTLR": None,
            "PREF_LINK": None,
            "RDN_EXCHD2": "DJI",
            "CONTEXT_ID": 2883,
            "DDS_DSO_ID": 8292,
            "SPS_SP_RIC": ".[SPSCMEIVAE1",
        },
    }
]
adc_response_with_error_peers = StubResponse(
    {
        "responses": [
            {
                "columnHeadersCount": 1,
                "data": [["peers(IBM.NNN)", None]],
                "error": [
                    {
                        "code": 413,
                        "col": 1,
                        "message": "Unable to resolve some identifier(s).",
                        "row": 0,
                    }
                ],
                "headerOrientation": "horizontal",
                "headers": [
                    [
                        {"displayName": "Instrument"},
                        {"displayName": "RIC", "field": "TR.RIC"},
                    ]
                ],
                "rowHeadersCount": 1,
                "totalColumnsCount": 2,
                "totalRowsCount": 2,
            }
        ]
    }
)
chains_response = {
    "meta": {
        "nodesLimit": "1",
        "nextLink": "",
        "prevLink": "",
        "cacheStatus": "",
    },
    "universe": {
        "ric": "0#.DJI",
        "displayName": "DJ INDU AVERAGE",
        "serviceName": "ELEKTRON_DD",
    },
    "data": {
        "constituents": [
            ".DJI",
            "AAPL.OQ",
            "AMGN.OQ",
            "AXP.N",
            "BA.N",
            "CAT.N",
            "CRM.N",
            "CSCO.OQ",
            "CVX.N",
            "DIS.N",
            "DOW.N",
            "GS.N",
            "HD.N",
            "HON.OQ",
        ]
    },
}

stream_data_with_error = [
    {
        "ID": 5,
        "Type": "Status",
        "Key": {"Service": "IDN_RDFNTS_CF", "Name": "IBM.NNN"},
        "State": {
            "Stream": "Closed",
            "Data": "Suspect",
            "Code": "NotFound",
            "Text": "The record could not be found",
        },
    }
]
constituents = [
    "AAPL.OQ",
    "AMGN.OQ",
    "AXP.N",
    "BA.N",
    "CAT.N",
    "CRM.N",
    "CSCO.OQ",
    "CVX.N",
    "DIS.N",
    "DOW.N",
    "GS.N",
    "HD.N",
    "HON.OQ",
]

summary_links = [".DJI"]


def make_http_request(adc_response=None, chain_response=None):
    def _http_request(request, *args, **kwargs):
        if request.url.endswith("streaming/pricing/v1/"):
            return StubResponse(content_data=infos)
        elif request.url == "test_get_rdp_url_root/data/datagrid/beta1/":
            return StubResponse(content_data=adc_response)
        elif request.url.startswith(
            "test_get_rdp_url_root/data/pricing/chains/v1/?universe="
        ):
            return StubResponse(content_data=chain_response)
        return StubResponse()

    return _http_request


def test_peers():
    session = StubSession(is_open=True, response=valid_adc_response)
    set_default(session)
    test_expression = "test_expr"
    peers = Peers(test_expression)
    assert peers.expression == f"peers({test_expression})"
    assert list(peers) == universe


def test_peers_exception():
    session = StubSession(is_open=True, response=adc_response_with_error_peers)
    set_default(session)
    test_expression = "test_expr"
    with pytest.raises(
        RDError,
        match="Error code 413 | Unable to resolve some identifier(s). Requested universes: peers(test_expr). Requested fields: TR.RIC",
    ):
        list(Peers(test_expression))


def test_screener():
    session = StubSession(is_open=True, response=valid_adc_response)
    set_default(session)
    test_expression = "test_expr"
    screener = Screener(test_expression)
    assert screener.expression == f"screen({test_expression})"
    assert list(screener) == universe


def test_screener_exception():
    session = StubSession(is_open=True, response=adc_response_with_error_screen)
    set_default(session)
    test_expression = "test_expr"
    with pytest.raises(
        RDError,
        match="Error code 800 | SCREEN(IBM.NNN) processing failed. Requested universes: screen(IBM.NNN). Requested fields: TR.RIC",
    ):
        list(Screener(test_expression))


@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_chain():
    session = StubSession(is_open=True, stream_auto_reconnection=True)

    session.http_request = make_http_request()
    send_ws_messages(session, stream_data)

    test_name = "Test_chain"
    result = Chain(test_name)

    assert result.constituents == constituents
    assert result.summary_links == summary_links
    assert result.name == test_name
    assert list(result) == constituents

    session.close()


adc_response = {
    "columnHeadersCount": 1,
    "data": [
        ["GS.N", "GS.N"],
        ["NKE.N", "NKE.N"],
        ["CSCO.OQ", "CSCO.OQ"],
        ["JPM.N", "JPM.N"],
        ["DIS.N", "DIS.N"],
        ["INTC.OQ", "INTC.OQ"],
        ["DOW.N", "DOW.N"],
        ["MRK.N", "MRK.N"],
        ["CVX.N", "CVX.N"],
        ["AXP.N", "AXP.N"],
        ["VZ.N", "VZ.N"],
        ["HD.N", "HD.N"],
        ["WBA.OQ", "WBA.OQ"],
        ["MCD.N", "MCD.N"],
        ["UNH.N", "UNH.N"],
        ["KO.N", "KO.N"],
        ["JNJ.N", "JNJ.N"],
        ["MSFT.OQ", "MSFT.OQ"],
        ["HON.OQ", "HON.OQ"],
        ["CRM.N", "CRM.N"],
        ["PG.N", "PG.N"],
        ["IBM.N", "IBM.N"],
        ["MMM.N", "MMM.N"],
        ["AAPL.OQ", "AAPL.OQ"],
        ["WMT.N", "WMT.N"],
        ["CAT.N", "CAT.N"],
        ["AMGN.OQ", "AMGN.OQ"],
        ["V.N", "V.N"],
        ["TRV.N", "TRV.N"],
        ["BA.N", "BA.N"],
    ],
    "headerOrientation": "horizontal",
    "headers": [
        [{"displayName": "Instrument"}, {"displayName": "RIC", "field": "TR.RIC"}]
    ],
    "rowHeadersCount": 1,
    "totalColumnsCount": 2,
    "totalRowsCount": 31,
}


@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_chain_exception():
    session = StubSession(is_open=True, stream_auto_reconnection=True)

    session.http_request = make_http_request()
    send_ws_messages(session, stream_data_with_error)

    test_name = "IBM.NNN"
    with pytest.raises(RDError, match="Error code -1 | The record could not be found"):
        list(Chain(test_name))

    session.close()


def scope_error(*args, **kwargs):
    raise ScopeError([{""}], {""}, "", "")


@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_chain_adc():
    session = StubSession(is_open=True, stream_auto_reconnection=True)
    session.config.set_param("apis.data.datagrid.underlying-platform", "rdp")

    session.http_request = make_http_request(adc_response)
    send_ws_messages(session, stream_data_with_error)

    test_name = "IBM.NNN"
    with mock.patch(
        "refinitiv.data.discovery._universe_expanders._chain.get_chain_data_from_stream",
        new=scope_error,
    ):
        assert list(Chain(test_name))

    session.close()


@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_chain_endpoint():
    session = StubSession(is_open=True, stream_auto_reconnection=True)
    session.config.set_param("apis.data.datagrid.underlying-platform", "rdp")

    session.http_request = make_http_request([], chains_response)
    send_ws_messages(session, stream_data_with_error)

    test_name = "IBM.NNN"
    with mock.patch(
        "refinitiv.data.discovery._universe_expanders._chain.get_chain_data_from_stream",
        new=scope_error,
    ):
        with mock.patch(
            "refinitiv.data.discovery._universe_expanders._chain.get_chain_data_from_adc",
            new=scope_error,
        ):
            assert list(Chain(test_name))

    session.close()


@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_chain_endpoint_with_bunch_constituents():
    session = StubSession(
        is_open=True, stream_auto_reconnection=True, response=BUNCH_CONSTITUENTS
    )
    session.config.set_param("apis.data.datagrid.underlying-platform", "rdp")

    send_ws_messages(session, stream_data_with_error)

    test_name = "0#.FCHI"
    with mock.patch(
        "refinitiv.data.discovery._universe_expanders._chain.get_chain_data_from_stream",
        new=scope_error,
    ):
        with mock.patch(
            "refinitiv.data.discovery._universe_expanders._chain.get_chain_data_from_adc",
            new=scope_error,
        ):
            testing_result = Chain(test_name)
            assert testing_result.constituents == EXPECTED_CONSTITUENTS_LIST
            assert testing_result.summary_links == EXPECTED_SUMMARY_LIST

    session.close()


@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_chain_endpoint_error():
    response = StubResponse({}, 401)
    response.errors = [RDError(401, "Not Authorized to Chain Data")]
    session = StubSession(
        is_open=True, stream_auto_reconnection=True, response=response
    )
    session.config.set_param("apis.data.datagrid.underlying-platform", "rdp")

    send_ws_messages(session, stream_data_with_error)

    test_name = "IBM.NNN"
    with mock.patch(
        "refinitiv.data.discovery._universe_expanders._chain.get_chain_data_from_stream",
        new=scope_error,
    ):
        with mock.patch(
            "refinitiv.data.discovery._universe_expanders._chain.get_chain_data_from_adc",
            new=scope_error,
        ):
            with pytest.raises(RDError):
                list(Chain(test_name))

    session.close()



import pytest

import refinitiv.data as rd
from refinitiv.data._errors import RDError
from refinitiv.data.delivery import cfs
from refinitiv.data.delivery.cfs import packages
from tests.unit.conftest import StubSession, StubResponse


@pytest.mark.parametrize(
    "value",
    [
        "get_data",
        "get_data_async",
    ],
)
def test_attribute(value):
    # given

    # when
    definition = packages.Definition("")

    # then
    assert hasattr(definition, value)


def test_invalid_fileset_id():
    session = StubSession(
        is_open=True,
        response=StubResponse(
            {
                "error": {
                    "id": "716e6298-24db-456a-b30f-2995f4c0ae84",
                    "code": "EDS.CFS.NoFilesetFound",
                    "status": 404,
                    "message": "No resource found for: ResourceType=Fileset; ResourceId=invalid fileset id",
                }
            },
            status_code=404,
        ),
    )
    rd.session.set_default(session)

    with pytest.raises(
        RDError,
        match=r"Error code 404 | No resource found for: ResourceType=Fileset; ResourceId=invalid fileset id",
    ):
        cfs.files.Definition("invalid fileset id").get_data()

    rd.session.set_default(None)


def test_non_existing_bucket_name():
    session = StubSession(
        is_open=True,
        response=StubResponse(
            {
                "error": {
                    "id": "53a7dae3-e455-483b-98d1-4500ce067065",
                    "code": "EDS.CFS.NoBucketFound",
                    "status": 404,
                    "message": "No resource found for: ResourceType=Bucket; ResourceId=non-existing bucket name",
                }
            },
            status_code=404,
        ),
    )
    rd.session.set_default(session)
    with pytest.raises(
        RDError,
        match=r"Error code 404 | No resource found for: ResourceType=Bucket; ResourceId=non-existing bucket name",
    ):
        cfs.file_sets.Definition(bucket="non-existing bucket name").get_data()

    rd.session.set_default(None)



from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    OverrideBidAsk,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    OverrideBidAskFields,
)


def test_override_bid_ask():
    # given
    expected_dict = {"instrumentCode": "string", "fields": {}, "date": "string"}

    # when
    testing_obj = OverrideBidAsk(
        instrument_code="string",
        fields=OverrideBidAskFields(),
        date="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.instrument_code == "string"
    assert testing_obj.fields == OverrideBidAskFields()
    assert testing_obj.date == "string"



import types
from functools import partial

import pytest

import refinitiv.data.content as rdc
import refinitiv.data.content.custom_instruments._enums
from refinitiv.data.content._intervals import Intervals
from refinitiv.data.content.estimates import Package
from refinitiv.data.content.historical_pricing import (
    EventTypes,
    Adjustments,
    MarketSession,
)
from refinitiv.data.content.ipa._enums import Direction, InterestType
from refinitiv.data.content.ipa._enums._asset_class import AssetClass
from refinitiv.data.content.ipa._enums._risk_type import RiskType
from refinitiv.data.content.news._sort_order import SortOrder as NewsSortOrder
from refinitiv.data.content.ownership import StatTypes
from refinitiv.data.content.search import Views
from refinitiv.data.content.symbol_conversion import SymbolTypes, CountryCode
from refinitiv.data.delivery._data._request import Request
from tests.unit.conftest import StubSession, StubResponse


def get_query_params_from_url(request: Request):
    url = request.url or ""
    query = url.split("?")[-1]
    query_params = [i.split("=") for i in query.split("&")]
    return dict(query_params)


def get_ipa_params(request: Request):
    universe = request.json.get("universe")
    return universe[0]


def get_ipa_instrument_definition_params(request: Request):
    universe = request.json.get("universe")
    return universe[0]["instrumentDefinition"]


def get_fund_params(request: Request):
    json = request.json
    entity = json.get("Entity")
    requests = entity.get("W", {}).get("requests", [{}])
    request = requests[0]
    rows = request.get("layout", {}).get("rows")
    item = {"item": "date"}
    if item in rows:
        return item
    else:
        assert False, f"{item} not in {rows}"


def get_json(request: Request):
    return request.json


def make_checker(param_name, value, func_get_params):
    def http_request(request: Request):
        params = func_get_params(request)
        assert str(params[param_name]) == str(value), (
            f"Param{param_name}:\n"
            f"Expected value {params[param_name]}. \n"
            f"Value {str(value)}"
        )
        return StubResponse()

    return http_request


def get_data(definition, params, session):
    try:
        if (
            isinstance(definition, types.FunctionType)
            or isinstance(definition, partial)
            and isinstance(definition.func, types.FunctionType)
        ):
            definition(**params, session=session)
        else:
            definition(**params).get_data(session=session)
    except (KeyError, IndexError):
        pass


def get_news_path(request: Request):
    url = request.url or ""
    view = url.rsplit("/", 1)[-1]
    result = {"view": view}
    return result


@pytest.mark.parametrize(
    ("definition", "enum_argname", "request_param_name", "enum", "func_get_params"),
    [
        (
            partial(rdc.estimates.view_actuals.annual.Definition, universe=""),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(rdc.estimates.view_actuals.interim.Definition, universe=""),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(rdc.estimates.view_summary.annual.Definition, universe=""),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(
                rdc.estimates.view_summary.historical_snapshots_non_periodic_measures.Definition,
                universe="",
            ),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(
                rdc.estimates.view_summary.historical_snapshots_periodic_measures_annual.Definition,
                universe="",
            ),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(
                rdc.estimates.view_summary.historical_snapshots_periodic_measures_interim.Definition,
                universe="",
            ),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(
                rdc.estimates.view_summary.historical_snapshots_recommendations.Definition,
                universe="",
            ),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(rdc.estimates.view_summary.interim.Definition, universe=""),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(
                rdc.estimates.view_summary.non_periodic_measures.Definition, universe=""
            ),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(rdc.estimates.view_summary.recommendations.Definition, universe=""),
            "package",
            "package",
            Package,
            get_query_params_from_url,
        ),
        (
            partial(rdc.fundamental_and_reference.Definition, universe="", fields=""),
            "row_headers",
            "item",
            rdc.fundamental_and_reference.RowHeaders,
            get_fund_params,
        ),
        (
            partial(rdc.historical_pricing.events.Definition, universe="VOD.L"),
            "eventTypes",
            "eventTypes",
            EventTypes,
            get_query_params_from_url,
        ),
        (
            partial(rdc.historical_pricing.events.Definition, universe="VOD.L"),
            "adjustments",
            "adjustments",
            Adjustments,
            get_query_params_from_url,
        ),
        (
            partial(rdc.historical_pricing.summaries.Definition, universe="VOD.L"),
            "interval",
            "interval",
            Intervals,
            get_query_params_from_url,
        ),
        (
            partial(rdc.historical_pricing.summaries.Definition, universe="VOD.L"),
            "adjustments",
            "adjustments",
            Adjustments,
            get_query_params_from_url,
        ),
        (
            partial(rdc.historical_pricing.summaries.Definition, universe="VOD.L"),
            "sessions",
            "sessions",
            MarketSession,
            get_query_params_from_url,
        ),
        (
            rdc.ipa.curves.zc_curve_definitions.Definition,
            "main_constituent_asset_class",
            "mainConstituentAssetClass",
            AssetClass,
            get_ipa_params,
        ),
        (
            rdc.ipa.curves.zc_curve_definitions.Definition,
            "risk_type",
            "riskType",
            RiskType,
            get_ipa_params,
        ),
        (
            rdc.ipa.financial_contracts.bond.Definition,
            "direction",
            "direction",
            Direction,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.bond.Definition,
            "interest_type",
            "interestType",
            InterestType,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.bond.Definition,
            "index_reset_frequency",
            "indexResetFrequency",
            rdc.ipa.financial_contracts.bond.Frequency,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.bond.Definition,
            "adjust_interest_to_payment_date",
            "adjustInterestToPaymentDate",
            rdc.ipa._enums.AdjustInterestToPaymentDate,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.bond.Definition,
            "index_compounding_method",
            "indexCompoundingMethod",
            rdc.ipa._enums.IndexCompoundingMethod,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.bond.Definition,
            "stub_rule",
            "stubRule",
            rdc.ipa._enums.StubRule,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition,
            "interest_payment_frequency",
            "interestPaymentFrequency",
            rdc.ipa._enums.Frequency,
            get_ipa_params,
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition,
            "interest_calculation_method",
            "interestCalculationMethod",
            rdc.ipa._enums.DayCountBasis,
            get_ipa_params,
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition,
            "payment_business_day_convention",
            "paymentBusinessDayConvention",
            rdc.ipa._enums.BusinessDayConvention,
            get_ipa_params,
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition,
            "payment_roll_convention",
            "paymentRollConvention",
            rdc.ipa._enums.DateRollingConvention,
            get_ipa_params,
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition,
            "index_reset_frequency",
            "indexResetFrequency",
            rdc.ipa.financial_contracts.cap_floor.Frequency,
            get_ipa_params,
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition,
            "index_reset_type",
            "indexResetType",
            rdc.ipa.financial_contracts.cap_floor.IndexResetType,
            get_ipa_params,
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition,
            "adjust_interest_to_payment_date",
            "adjustInterestToPaymentDate",
            rdc.ipa._enums.AdjustInterestToPaymentDate,
            get_ipa_params,
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition,
            "stub_rule",
            "stubRule",
            rdc.ipa._enums.StubRule,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition,
            "buy_sell",
            "buySell",
            rdc.ipa._enums.BuySell,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.cds.Definition,
            "cds_convention",
            "cdsConvention",
            rdc.ipa._enums.CdsConvention,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.cds.Definition,
            "start_date_moving_convention",
            "startDateMovingConvention",
            rdc.ipa._enums.BusinessDayConvention,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.cds.Definition,
            "end_date_moving_convention",
            "endDateMovingConvention",
            rdc.ipa._enums.BusinessDayConvention,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.cross.Definition,
            "fx_cross_type",
            "fxCrossType",
            rdc.ipa._enums.FxCrossType,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.option.Definition,
            "buy_sell",
            "buySell",
            rdc.ipa._enums.BuySell,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.option.Definition,
            "call_put",
            "callPut",
            rdc.ipa._enums.CallPut,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.option.Definition,
            "exercise_style",
            "exerciseStyle",
            rdc.ipa._enums.ExerciseStyle,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.option.Definition,
            "underlying_type",
            "underlyingType",
            rdc.ipa.financial_contracts.option._enums.UnderlyingType,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.repo.Definition,
            "day_count_basis",
            "dayCountBasis",
            rdc.ipa._enums.DayCountBasis,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.swaption.Definition,
            "buy_sell",
            "buySell",
            rdc.ipa._enums.BuySell,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.swaption.Definition,
            "exercise_style",
            "exerciseStyle",
            rdc.ipa.financial_contracts.swaption.ExerciseStyle,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.swaption.Definition,
            "premium_settlement_type",
            "premiumSettlementType",
            rdc.ipa.financial_contracts.swaption.PremiumSettlementType,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.swaption.Definition,
            "settlement_type",
            "settlementType",
            rdc.ipa.financial_contracts.swaption.SwaptionSettlementType,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.swaption.Definition,
            "swaption_type",
            "swaptionType",
            rdc.ipa.financial_contracts.swaption.SwaptionType,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.term_deposit.Definition,
            "payment_business_day_convention",
            "paymentBusinessDayConvention",
            rdc.ipa._enums.BusinessDayConvention,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.term_deposit.Definition,
            "payment_roll_convention",
            "paymentRollConvention",
            rdc.ipa._enums.DateRollingConvention,
            get_ipa_instrument_definition_params,
        ),
        (
            rdc.ipa.financial_contracts.term_deposit.Definition,
            "year_basis",
            "yearBasis",
            rdc.ipa._enums.DayCountBasis,
            get_ipa_instrument_definition_params,
        ),
        (
            partial(rdc.news.headlines.Definition, query=""),
            "sort_order",
            "sortOrder",
            NewsSortOrder,
            get_query_params_from_url,
        ),
        (
            partial(rdc.ownership.consolidated.breakdown.Definition, universe=""),
            "stat_type",
            "statType",
            StatTypes,
            get_query_params_from_url,
        ),
        (
            partial(rdc.ownership.consolidated.recent_activity.Definition, universe=""),
            "sort_order",
            "sortOrder",
            rdc.ownership.SortOrder,
            get_query_params_from_url,
        ),
        (
            partial(
                rdc.ownership.consolidated.shareholders_history_report.Definition,
                universe="",
            ),
            "frequency",
            "frequency",
            rdc.ownership.Frequency,
            get_query_params_from_url,
        ),
        (
            partial(rdc.ownership.fund.breakdown.Definition, universe=""),
            "stat_type",
            "statType",
            StatTypes,
            get_query_params_from_url,
        ),
        (
            partial(rdc.ownership.fund.recent_activity.Definition, universe=""),
            "sort_order",
            "sortOrder",
            rdc.ownership.SortOrder,
            get_query_params_from_url,
        ),
        (
            partial(
                rdc.ownership.fund.shareholders_history_report.Definition, universe=""
            ),
            "frequency",
            "frequency",
            rdc.ownership.Frequency,
            get_query_params_from_url,
        ),
        (
            partial(rdc.symbol_conversion.Definition, symbols=""),
            "from_symbol_type",
            "fromSymbolType",
            SymbolTypes,
            get_json,
        ),
        (
            partial(rdc.symbol_conversion.Definition, symbols=""),
            "to_symbol_types",
            "toSymbolType",
            SymbolTypes,
            get_json,
        ),
        (
            partial(rdc.symbol_conversion.Definition, symbols=""),
            "preferred_country_code",
            "preferredCountryCode",
            CountryCode,
            get_json,
        ),
        (
            partial(rdc.symbol_conversion.Definition, symbols=""),
            "asset_class",
            "assetClass",
            rdc.symbol_conversion.AssetClass,
            get_json,
        ),
        (
            partial(rdc.symbol_conversion.Definition, symbols=""),
            "asset_state",
            "assetState",
            rdc.symbol_conversion.AssetState,
            get_json,
        ),
        # deprecated method
        # (
        #     partial(rdc.custom_instruments.manage.create, symbol="S)Name.GE-0000"),
        #     "type_",
        #     "type",
        #     refinitiv.data.content.custom_instruments._enums.CustomInstrumentTypes,
        #     get_json,
        # ),
    ],
)
def test_enums_definitions(
    definition, enum_argname, request_param_name, enum, func_get_params
):
    session = StubSession(is_open=True)
    session.config.set_param("apis.data.news.underlying-platform", "rdp")
    for enum_item in enum:
        session.http_request = make_checker(
            request_param_name, enum_item.value, func_get_params
        )
        get_data(definition, {enum_argname: enum_item}, session)
        value = enum_item.value
        get_data(definition, {enum_argname: value}, session)
        if isinstance(value, str):
            get_data(definition, {enum_argname: value.lower()}, session)
            get_data(definition, {enum_argname: value.upper()}, session)

    # invalid
    for invalid in {111, "INVALID"}:
        try:
            get_data(definition, {enum_argname: invalid}, session)
        except TypeError as e:
            try:
                assert f"Parameter '{enum_argname}' of invalid type provided:" in str(e)
            except AssertionError:
                try:
                    assert (
                        f"Parameter '{request_param_name}' of invalid type provided:"
                        in str(e)
                    )
                except AssertionError as e:
                    pytest.skip(str(e))
        except AttributeError as e:
            assert f"Value '{invalid}' must be in " in str(e)
        except ValueError as e:
            assert f"Cannot convert param '{invalid}'" == str(e)
        else:
            assert False


@pytest.mark.parametrize(
    ("definition", "enum_argname", "request_param_name", "enum"),
    [
        (
            rdc.trade_data_service.Definition,
            "universe_type",
            "universeType",
            rdc.trade_data_service.UniverseTypes,
        ),
        (
            rdc.trade_data_service.Definition,
            "events",
            "events",
            rdc.trade_data_service.Events,
        ),
        (
            rdc.trade_data_service.Definition,
            "finalized_orders",
            "finalizedOrders",
            rdc.trade_data_service.FinalizedOrders,
        ),
    ],
)
def test_enums_in_definitions_for_stream(
    definition, enum_argname, request_param_name, enum
):
    session = StubSession(is_open=True)
    for enum_item in enum:
        stream = definition(**{enum_argname: enum_item}).get_stream(session)
        open_message = stream._stream._stream.open_message
        enum_item_value = enum_item.value
        assert open_message["parameters"][request_param_name] == enum_item_value
        stream = definition(**{enum_argname: enum_item_value}).get_stream(session)
        open_message = stream._stream._stream.open_message
        assert open_message["parameters"][request_param_name] == enum_item_value

    with pytest.raises(AttributeError, match="Value '123' must be in"):
        definition(**{enum_argname: 123})
    with pytest.raises(AttributeError, match="Value 'INVALID___' must be in"):
        definition(**{enum_argname: "INVALID___"})



__all__ = (
    "Definition",
    "ButterflyShift",
    "CombinedShift",
    "CrossCurrencyCurveDefinitionPricing",
    "FlatteningShift",
    "FxForwardCurveDefinition",
    "FxForwardCurveParameters",
    "FxShiftScenario",
    "LongEndShift",
    "ParallelShift",
    "ShiftDefinition",
    "ShortEndShift",
    "TimeBucketShift",
    "TwistShift",
    "ValuationTime",
    "ArrayMainConstituentAssetClass",
    "ArrayRiskType",
    "InterpolationMode",
    "ConstituentOverrideMode",
    "MainConstituentAssetClass",
    "RiskType",
    "ShiftType",
    "ShiftUnit",
)


from ._definition import Definition
from ...._curves._cross_currency_curves._curves import (
    ButterflyShift,
    CombinedShift,
    CrossCurrencyCurveDefinitionPricing,
    FlatteningShift,
    FxForwardCurveDefinition,
    FxForwardCurveParameters,
    FxShiftScenario,
    LongEndShift,
    ParallelShift,
    ShiftDefinition,
    ShortEndShift,
    TimeBucketShift,
    TwistShift,
    ValuationTime,
)

from ._enums import (
    ArrayMainConstituentAssetClass,
    ArrayRiskType,
    InterpolationMode,
    ConstituentOverrideMode,
    MainConstituentAssetClass,
    RiskType,
    ShiftType,
    ShiftUnit,
)



# coding: utf8
from ..._object_definition import ObjectDefinition
from ....._types import OptStr


class Step(ObjectDefinition):
    """
    Parameters
    ----------
    effective_date : str, optional

    meeting_date : str, optional

    """

    def __init__(
        self,
        effective_date: OptStr = None,
        meeting_date: OptStr = None,
    ) -> None:
        super().__init__()
        self.effective_date = effective_date
        self.meeting_date = meeting_date

    @property
    def effective_date(self):
        """
        :return: str
        """
        return self._get_parameter("effectiveDate")

    @effective_date.setter
    def effective_date(self, value):
        self._set_parameter("effectiveDate", value)

    @property
    def meeting_date(self):
        """
        :return: str
        """
        return self._get_parameter("meetingDate")

    @meeting_date.setter
    def meeting_date(self, value):
        self._set_parameter("meetingDate", value)



import json

from refinitiv.data.content import news
from tests.unit.conftest import StubSession, StubResponse
from tests.unit.content.news.top_news.hierarchy.data_for_tests import (
    top_news_hierarchy_df,
)

with open(
    "tests/unit/content/news/top_news/hierarchy/top-news-hierarchy.json", "r"
) as f:
    top_news_response = json.loads(f.read())


def test_top_news_hierarchy_dataframe():
    session = StubSession(is_open=True, response=StubResponse(top_news_response))

    response = news.top_news.hierarchy.Definition().get_data(session=session)

    assert response.data.df.to_string() == top_news_hierarchy_df
    assert response.data.hierarchy



import asyncio

import pytest

from refinitiv.data.content._content_provider_layer import ContentProviderLayer
from tests.unit.conftest import StubSession
from tests.unit.content.conftest import StubDefinition


def test_content_provider_layer_get_data(content_type):
    # given
    session = StubSession(is_open=True)
    provider_layer = ContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # when
    data = provider_layer.get_data(session)

    # then
    assert data


@pytest.mark.asyncio
async def test_content_provider_layer_get_data_async(content_type):
    # given
    session = StubSession(is_open=True)
    provider_layer = ContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # when
    data = await provider_layer.get_data_async(session)

    # then
    assert data


@pytest.mark.asyncio
async def test_content_provider_layer_get_data_pass_on_response(content_type):
    # given
    fut = asyncio.Future()

    def on_response(response, provider_layer, session):
        fut.set_result(response)

    session = StubSession(is_open=True)
    provider_layer = ContentProviderLayer(
        content_type,
        definition=StubDefinition(),
        universe=StubDefinition(),
    )

    # when
    provider_layer.get_data(session, on_response=on_response)

    response = await fut

    # then
    assert response


@pytest.mark.asyncio
async def test_content_provider_layer_call_get_data_async_pass_on_response(
    content_type,
):
    # given
    fut = asyncio.Future()

    def on_response(response, provider_layer, session):
        fut.set_result(response)

    session = StubSession(is_open=True)
    provider_layer = ContentProviderLayer(
        content_type,
        definition=StubDefinition(),
        universe=StubDefinition(),
    )

    # when
    await provider_layer.get_data_async(session, on_response=on_response)

    response = await fut

    # then
    assert response


def test_content_provider_layer_call_get_data_will_raise_error_is_session_closed(
    content_type,
):
    # given
    session = StubSession(is_open=False)
    provider_layer = ContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # then
    with pytest.raises(Exception, match="Session is not opened."):
        # when
        provider_layer.get_data(session)


@pytest.mark.asyncio
async def test_content_provider_layer_call_get_data_async_will_raise_error_is_session_closed(
    content_type,
):
    # given
    session = StubSession(is_open=False)
    provider_layer = ContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # then
    with pytest.raises(Exception, match="Session is not opened."):
        # when
        await provider_layer.get_data_async(session)
